import logging
import os
import sys
from typing import Callable, Optional

from vistautils.logging_utils import configure_logging_from
from vistautils.parameters import Parameters, YAMLParametersLoader

log = logging.getLogger(__name__)  # pylint:disable=invalid-name


def parameters_only_entry_point(
    main_method: Callable[[Parameters], None],
    usage_message: str = None,
    *,
    parameters: Optional[Parameters] = None
) -> None:
    """
    Convenience wrapper for entry points which take a single parameter file as an argument.

    In addition to saving the boilerplate of loading parameters, this will also automatically
    configure logging from the param file itself (see `configure_logging_from`) and log the
    contents of the parameter file. In the future, other such conveniences may be added.

    The user may specify *parameters* explicitly,
    in which case the argument passed to the program is ignored.

    This is primarily for ISI VISTA-internal use.
    """
    if len(sys.argv) == 2:
        if parameters:
            params = parameters
        else:
            params = YAMLParametersLoader().load(sys.argv[1])
        configure_logging_from(params)
        log.info("Ran with parameters:\n%s", params)
        main_method(params)
    else:
        if not usage_message:
            import __main__ as main

            usage_message = "usage: {!s} param_file".format(
                os.path.basename(main.__file__)
            )
        log.error(usage_message)
        sys.exit(1)



#!/usr/bin/env python

"""
Split any character key-value store in multiple character key-value stores.

This is useful when splitting up input for parallel processing.

Currently the output stores are always zip file stores. This might become configurable
in the future.

The "input" namespace specifies the input key-value store. Please see
char_key_value_source_from_params for details on the parameters used to specify input.

The "num_slices" param specifies how many slices to create.

The list of zip files created will be stored in "_slices.txt" in the output directory.
"""
from contextlib import ExitStack

from vistautils.io_utils import CharSink
from vistautils.key_value import KeyValueSink, char_key_value_linear_source_from_params
from vistautils.parameters import Parameters
from vistautils.parameters_only_entrypoint import parameters_only_entry_point


def main(params: Parameters):
    output_directory = params.creatable_directory("output_dir")
    slices = params.positive_integer("num_slices")

    slice_paths = [output_directory / "{!s}.zip".format(i) for i in range(slices)]
    CharSink.to_file(output_directory / "_slices.txt").write(
        "\n".join(str(x) for x in slice_paths)
    )
    output_sinks = [
        KeyValueSink.zip_character_sink(slice_path) for slice_path in slice_paths
    ]

    # this is the magic incantation for handling variable-length lists of context managers
    with ExitStack() as exit_stack:
        for output_sink in output_sinks:
            exit_stack.enter_context(output_sink)
        with char_key_value_linear_source_from_params(params) as input_source:
            for (i, v) in enumerate(input_source.items()):
                output_sinks[i % slices].put(v[0], v[1])


if __name__ == "__main__":
    parameters_only_entry_point(main)



# really needs to be totally-ordered
import warnings
from abc import ABCMeta, abstractmethod
from typing import (
    Any,
    Container,
    Generic,
    Hashable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Sized,
    Tuple,
    TypeVar,
    Union,
)

from attr import attrib, attrs, validators
from immutablecollections import ImmutableDict, ImmutableSet, immutabledict, immutableset
from sortedcontainers import SortedDict

# Port of Guava's Range data type and associated classes
from vistautils.preconditions import check_arg, check_not_none

# will be initialized after bound type declarations
# noinspection PyTypeHints
_OPEN: "BoundType" = None  # type:ignore
# noinspection PyTypeHints
_CLOSED: "BoundType" = None  # type:ignore


class BoundType:
    """
    A possible type of boundary for a range.

    A boundary is either closed, meaning it includes its endpoint, or open, meaning it does not.
    """

    __slots__ = ()

    @staticmethod
    def open() -> "BoundType":
        return _OPEN

    @staticmethod
    def closed() -> "BoundType":
        return _CLOSED

    def flip(self):
        return _CLOSED if self is _OPEN else _OPEN


class _Open(BoundType):
    __slots__ = ()


class _Closed(BoundType):
    __slots__ = ()


# noinspection PyRedeclaration,PyTypeHints
_OPEN: BoundType = _Open()  # type: ignore
# noinspection PyRedeclaration,PyTypeHints
_CLOSED: BoundType = _Closed()  # type: ignore

# these need to be initialized after declaration of _Cut
# noinspection PyTypeHints
_BELOW_ALL: "_Cut" = None  # type:ignore
# noinspection PyTypeHints
_ABOVE_ALL: "_Cut" = None  # type:ignore

# T needs to be comparable, but Python typing seems to lack a way to specify this?
# see https://github.com/python/mypy/issues/500
# we track this with our own issue #201
T = TypeVar("T")


class _Cut(Generic[T], metaclass=ABCMeta):
    """
    Implementation detail for the internal structure of Range instances.

    Represents a unique way of "cutting" a "number line" into two sections; this can be done below
    a certain value, above a certain value, below all values or above all values.
    With this  object defined in this way, an interval can always be represented by a pair of
    Cut instances.

    This is a Python port of Guava code originally written by Kevin Bourrillion.
    """

    __slots__ = ()

    @property
    @abstractmethod
    def endpoint(self) -> T:
        raise NotImplementedError()

    @abstractmethod
    def is_less_than(self, other: T) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def as_upper_bound(self) -> BoundType:
        raise NotImplementedError()

    @abstractmethod
    def as_lower_bound(self) -> BoundType:
        raise NotImplementedError()

    @abstractmethod
    def describe_as_lower_bound(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def describe_as_upper_bound(self) -> str:
        raise NotImplementedError()

    def compare_to(self, other: "_Cut[T]") -> int:
        # overridden by BelowAll, AboveAll
        if other is _BELOW_ALL:
            return 1
        if other is _ABOVE_ALL:
            return -1
        if self.endpoint < other.endpoint:  # type: ignore
            return -1
        elif other.endpoint < self.endpoint:  # type: ignore
            return 1
        else:
            # BelowValue precedes AboveValue
            if isinstance(self, _AboveValue):
                if isinstance(other, _AboveValue):
                    return 0
                else:
                    return 1
            else:
                if isinstance(other, _AboveValue):
                    return -1
                else:
                    return 0

    # cannot use @totalordering because we want to compare across sub-classes
    # so we need to spell out all the operators
    def __lt__(self, other) -> bool:
        return self.compare_to(other) < 0

    def __le__(self, other) -> bool:
        return self.compare_to(other) <= 0

    def __gt__(self, other) -> bool:
        return self.compare_to(other) > 0

    def __ge__(self, other) -> bool:
        return self.compare_to(other) >= 0

    def __eq__(self, other) -> bool:
        return self.compare_to(other) == 0


@attrs(frozen=True, slots=True, hash=False, cmp=False)
class _BelowAll(_Cut[T]):
    # pylint:disable=protected-access
    @property
    def endpoint(self) -> T:
        raise ValueError("BelowAll cut lacks an endpoint")

    def is_less_than(self, other: T) -> bool:  # pylint:disable=unused-argument
        return True

    def as_upper_bound(self) -> BoundType:
        raise AssertionError("Should never be called")

    def as_lower_bound(self) -> BoundType:
        raise AssertionError("Should never be called")

    def describe_as_lower_bound(self) -> str:
        return "(-\u221e"  # Returns ( and a negative infinity

    def describe_as_upper_bound(self) -> str:
        raise AssertionError("Can't happen")

    def compare_to(self, other: "_Cut[T]") -> int:
        # we assume only the constant _BELOW_ALL is ever instantiated
        if other is self:
            return 0
        return -1

    def __hash__(self):
        # some arbitrary number
        return 233904909


@attrs(frozen=True, slots=True, hash=False, cmp=False)
class _AboveAll(_Cut[T]):
    # pylint:disable=protected-access
    @property
    def endpoint(self) -> T:
        raise ValueError("AboveAll cut lacks an endpoint")

    def is_less_than(self, other: T) -> bool:  # pylint:disable=unused-argument
        return False

    def as_upper_bound(self) -> BoundType:
        raise AssertionError("Should never be called")

    def as_lower_bound(self) -> BoundType:
        raise AssertionError("Should never be called")

    def describe_as_lower_bound(self) -> str:
        raise AssertionError("Can't happen")

    def describe_as_upper_bound(self) -> str:
        return "+\u221e)"  # Returns positive infinity and )

    def compare_to(self, other: "_Cut[T]") -> int:
        # we assume only the constant _ABOVE_ALL is ever instantiated
        if other is self:
            return 0
        return 1

    def __hash__(self):
        # some arbitrary number
        return 9989388


# noinspection PyRedeclaration
_BELOW_ALL = _BelowAll()
# noinspection PyRedeclaration
_ABOVE_ALL = _AboveAll()


@attrs(frozen=True, slots=True, repr=False, hash=False, cmp=False)
class _BelowValue(_Cut[T]):
    # pylint:disable=protected-access
    _endpoint = attrib()

    @property
    def endpoint(self) -> T:
        return self._endpoint

    def is_less_than(self, other: T) -> bool:
        return self._endpoint < other or self._endpoint == other

    def as_upper_bound(self) -> BoundType:
        return _OPEN

    def as_lower_bound(self) -> BoundType:
        return _CLOSED

    def __hash__(self):
        return hash(self._endpoint)

    def __eq__(self, other):
        if isinstance(other, _BelowValue):
            return self._endpoint == other._endpoint
        return False

    def describe_as_lower_bound(self) -> str:
        return "[%s" % self._endpoint

    def describe_as_upper_bound(self) -> str:
        return "%s)" % self._endpoint

    def __repr__(self) -> str:
        return "\\\\%s/" % self._endpoint


@attrs(frozen=True, slots=True, repr=False, hash=False, cmp=False)
class _AboveValue(_Cut[T]):
    # pylint:disable=protected-access
    _endpoint = attrib()

    @property
    def endpoint(self) -> T:
        return self._endpoint

    def is_less_than(self, other: T) -> bool:
        return self._endpoint < other

    def as_upper_bound(self) -> BoundType:
        return _CLOSED

    def as_lower_bound(self) -> BoundType:
        return _OPEN

    def __hash__(self):
        # bitwise complement to distinguish it from the corresponding _BelowValue
        return ~hash(self._endpoint)

    def __eq__(self, other):
        if isinstance(other, _AboveValue):
            return self._endpoint == other._endpoint
        return False

    def describe_as_lower_bound(self) -> str:
        return "(%s" % self._endpoint

    def describe_as_upper_bound(self) -> str:
        return "%s]" % self._endpoint

    def __repr__(self) -> str:
        return "/%s\\\\" % self._endpoint


# must initialize after declaring Range
# noinspection PyTypeHints
RANGE_ALL: "Range" = None  # type: ignore


# this should have slots=True but cannot for the moment due to
# https://github.com/python-attrs/attrs/issues/313
# Pylint disable due to https://github.com/PyCQA/pylint/issues/2472
@attrs(
    frozen=True, repr=False, cmp=False, hash=False
)  # pylint: disable=inherit-non-class
class Range(Container[T], Generic[T], Hashable):
    """
    The boundaries of a contiguous span of values.

    The value must be of some type which implements `<` in a way consistent with `__eq__`.
    Note this does not provide a means of iterating over these values.

    Each end of the `Range` may be *bounded* or *unbounded*. If bounded, there is an associated
     *endpoint* value and the range is considered either *open* (does not include the endpoint
     value) or *closed* (does include the endpoint value).  With three possibilities on each
     side, this yields nine basic types of ranges, enumerated belows.
     (Notation: a square bracket (`[ ]`) indicates that the range is closed on that side;
     a parenthesis (`( )`) means it is either open or unbounded. The construct `{x | statement}`
      is read "the set of all x such that statement.")

    ========        ==========            ==============
    Notation	    Definition	          Factory method
    ========        ==========            ==============
    `(a..b)`	    `{x | a < x < b}`	      open
    `[a..b]`	    `{x | a <= x <= b}`	      closed
    `(a..b]`	    `{x | a < x <= b}`	      open_closed
    `[a..b)`	    `{x | a <= x < b}`	      closed_open
    `(a..+∞)`	    `{x | x > a}`	          greater_than
    `[a..+∞)`	    `{x | x >= a}`	          at_least
    `(-∞..b)`	    `{x | x < b}`	          less_than
    `(-∞..b]`	    `{x | x <= b}`	          at_most
    `(-∞..+∞)`	    `{x}`	                  all
    =========       ===========           ==============

    When both endpoints exist, the upper endpoint may not be less than the lower. The endpoints may
     be equal only if at least one of the bounds is closed:

     * `[a..a]` : a singleton range
     * `[a..a)`; `(a..a]` : empty ranges; also valid
     * `(a..a)` : invalid; an exception will be thrown

    ========
    Warnings
    ========

    * Use immutable value types only, if at all possible. If you must use a mutable type, do not
    allow the endpoint instances to mutate after the range is created!

    =======
    Notes
    ========
    * Instances of this type are obtained using the static factory methods in this class.
    * Ranges are convex: whenever two values are contained, all values in between them must also be
         contained. More formally, for any `c1 <= c2 <= c3` of type `C`,
         `c1 in r and c3 in r` implies `c2 in r`. This means that a `Range[int]` can never
         be used to represent, say, "all prime numbers from 1 to 100."
    * Terminology note: a range `a` is said to be the maximal range having property `P` if,
    for all ranges `b` also having property `P`, `a.encloses(b)`. Likewise, `a` is minimal when
    `b.encloses(a)` for all `b` having property `P`. See, for example, the definition of
    intersection.

    This class (including the documentation) is an almost direct translation of Guava's Range,
    which was originally authored by Kevin Bourrillion and Gregory Kick.
    """

    # pylint:disable=protected-access
    _lower_bound: _Cut[T] = attrib(validator=validators.instance_of(_Cut))
    _upper_bound: _Cut[T] = attrib(validator=validators.instance_of(_Cut))

    def __attrs_post_init__(self):
        check_arg(
            self._lower_bound <= self._upper_bound,
            "Upper bound of a range cannot be less than lower bound but got %s ",
            (self,),
        )
        check_arg(self._lower_bound is not _ABOVE_ALL)
        check_arg(self._upper_bound is not _BELOW_ALL)

    @staticmethod
    def open(lower: T, upper: T) -> "Range[T]":
        return Range(_AboveValue(lower), _BelowValue(upper))

    @staticmethod
    def closed(lower: T, upper: T) -> "Range[T]":
        return Range(_BelowValue(lower), _AboveValue(upper))

    @staticmethod
    def closed_open(lower: T, upper: T) -> "Range[T]":
        return Range(_BelowValue(lower), _BelowValue(upper))

    @staticmethod
    def open_closed(lower: T, upper: T) -> "Range[T]":
        return Range(_AboveValue(lower), _AboveValue(upper))

    @staticmethod
    def less_than(upper: T) -> "Range[T]":
        return Range(_BELOW_ALL, _BelowValue(upper))

    @staticmethod
    def at_most(upper: T) -> "Range[T]":
        return Range(_BELOW_ALL, _AboveValue(upper))

    @staticmethod
    def greater_than(lower: T) -> "Range[T]":
        return Range(_AboveValue(lower), _ABOVE_ALL)

    @staticmethod
    def at_least(lower: T) -> "Range[T]":
        return Range(_BelowValue(lower), _ABOVE_ALL)

    @staticmethod
    def all() -> "Range[T]":
        return RANGE_ALL

    @staticmethod
    def create_spanning(ranges: Sequence["Range[T]"]):
        if not ranges:
            raise ValueError("Cannot create range from span of empty range collection")
        return Range(
            min(x._lower_bound for x in ranges), max(x._upper_bound for x in ranges)
        )

    def has_lower_bound(self) -> bool:
        return self._lower_bound is not _BELOW_ALL

    def has_upper_bound(self) -> bool:
        return self._upper_bound is not _ABOVE_ALL

    @property
    def lower_bound_type(self) -> BoundType:
        return self._lower_bound.as_lower_bound()

    @property
    def upper_bound_type(self) -> BoundType:
        return self._upper_bound.as_upper_bound()

    @property
    def lower_endpoint(self) -> T:
        return self._lower_bound.endpoint

    @property
    def upper_endpoint(self) -> T:
        return self._upper_bound.endpoint

    def is_empty(self) -> bool:
        """
        Determine if a range is empty.

        Returns `True` if this range is of the form `[v..v)` or `(v..v]`.
        (This does not encompass ranges of the form (v..v), because such ranges are invalid and
        can't be constructed at all.)

        Note that certain discrete ranges such as the integer range (3..4) are not considered empty,
         even though they contain no actual values.
        """
        return self._lower_bound == self._upper_bound

    # I don't know why mypy complains about narrowing the type, which seems a reasonable thing to do
    def __contains__(self, val: T) -> bool:  # type: ignore
        check_not_none(val)
        return self._lower_bound.is_less_than(val) and not self._upper_bound.is_less_than(
            val
        )

    def encloses(self, other: "Range[T]") -> bool:
        # noinspection PyChainedComparisons
        return (
            self._lower_bound.compare_to(other._lower_bound) <= 0
            and self._upper_bound.compare_to(other._upper_bound) >= 0
        )

    def is_connected(self, other: "Range[T]") -> bool:
        """
        Determine if two ranges are connected.

        Returns `True` if there exists a (possibly empty) range which is enclosed by both this
        range and `other`. For example,

        * `[2, 4)` and `[5, 7)` are not connected
        * `[2, 4)` and `[3, 5)` are connected, because both enclose `[3, 4)`
        * `[2, 4)` and `[4, 6)` are connected, because both enclose the empty range `[4, 4)`

        Note that this range and `other` have a well-defined union and intersection (as a
        single, possibly-empty range) if and only if this method returns `True`.

        The connectedness relation is both reflexive and symmetric, but does not form an
        equivalence relation as it is not transitive.

        Note that certain discrete ranges are not considered connected, even though there are
        no elements "between them." For example, `[3, 5]` is not considered connected to
        `[6, 10]`.
        """
        return (
            self._lower_bound <= other._upper_bound
            and other._lower_bound <= self._upper_bound
        )

    def span(self, other: "Range[T]") -> "Range[T]":
        """
        Get the minimal range enclosing both this range and `other`.

         For example, the span of `[ 1..3] and (5..7)` is `[1..7)`.
        If the input ranges are connected, the returned range can also be called their union.
        If they are not, note that the span might contain values that are not contained in either
        input range.

        Like intersection, this operation is commutative, associative and idempotent. Unlike it, it
        is always well-defined for any two input ranges.
        """
        lower_cmp = self._lower_bound.compare_to(other._lower_bound)
        upper_cmp = self._upper_bound.compare_to(other._upper_bound)

        if lower_cmp <= 0 and upper_cmp >= 0:
            return self
        elif lower_cmp >= 0 and upper_cmp <= 0:
            return other
        else:
            return Range(
                self._lower_bound if lower_cmp <= 0 else other._lower_bound,
                self._upper_bound if upper_cmp >= 0 else other._upper_bound,
            )

    def intersection(self, connected_range: "Range[T]") -> "Range[T]":
        """
        Get the intersection of this range and `other`.

        Returns the maximal range enclosed by both this range and connectedRange, if such a
        range exists.

        For example, the intersection of `[1..5]` and `(3..7)` is `(3..5]`. The
        resulting range may be empty; for example, `[1..5)` intersected with `[5..7)`
        yields the empty range `[5..5)`.

        The intersection exists if and only if the two ranges are connected.  This method throws
        a `ValueError` is `connected_range` is not in fact connected.

        The intersection operation is commutative, associative and idempotent, and its identity
        element is the `all` range/
        """
        lower_cmp = self._lower_bound.compare_to(connected_range._lower_bound)
        upper_cmp = self._upper_bound.compare_to(connected_range._upper_bound)
        if lower_cmp >= 0 >= upper_cmp:
            return self
        elif lower_cmp <= 0 <= upper_cmp:
            return connected_range
        else:
            return Range(
                self._lower_bound if lower_cmp >= 0 else connected_range._lower_bound,
                self._upper_bound if upper_cmp <= 0 else connected_range._upper_bound,
            )

    def intersects(self, other_range: "Range[T]") -> bool:
        """
        Determine if this range i
        Args:
            other_range:

        Returns:

        """
        lower_cmp = self._lower_bound.compare_to(other_range._lower_bound)
        upper_cmp = self._upper_bound.compare_to(other_range._upper_bound)
        if lower_cmp >= 0 >= upper_cmp:
            return True
        elif lower_cmp <= 0 <= upper_cmp:
            return True
        else:
            intersection_lb = (
                self._lower_bound if lower_cmp >= 0 else other_range._lower_bound
            )
            intersection_ub = (
                self._upper_bound if upper_cmp <= 0 else other_range._upper_bound
            )
            return intersection_lb <= intersection_ub

    def __eq__(self, other) -> bool:
        if isinstance(other, Range):
            return (
                self._lower_bound == other._lower_bound
                and self._upper_bound == other._upper_bound
            )
        return False

    def __hash__(self) -> int:
        return hash(self._lower_bound) + 31 * hash(self._upper_bound)

    def __repr__(self) -> str:
        return (
            self._lower_bound.describe_as_lower_bound()
            + ".."
            + self._upper_bound.describe_as_upper_bound()
        )


# noinspection PyRedeclaration
RANGE_ALL = Range(_BELOW_ALL, _ABOVE_ALL)


# Pylint disable due to https://github.com/PyCQA/pylint/issues/2472
class RangeSet(
    Generic[T], Container[T], Sized, metaclass=ABCMeta
):  # pylint:disable=E0239
    """
    A set comprising zero or more nonempty, disconnected ranges of type `T`.

    Implementations that choose to support the `add(Range)` operation are required to ignore empty
    ranges and coalesce connected ranges. For example ::

         rangeSet: RangeSet[int] = TreeRangeSet();`
         rangeSet.add(Range.closed(1, 10)); // {[1, 10]}
         rangeSet.add(Range.closed_open(11, 15)); // disconnected range; {[1, 10], [11, 15)}
         rangeSet.add(Range.closed_open(15, 20)); // connected range; {[1, 10], [11, 20)}
         rangeSet.add(Range.open_closed(0, 0)); // empty range; {[1, 10], [11, 20)}
         rangeSet.remove(Range.open(5, 10)); // splits [1, 10]; {[1, 5], [10, 10], [11, 20)}


    Note that the behavior of `Range.isEmpty()` and `Range.isConnected(Range)` may not be as
    expected on discrete ranges. See the documentation of those methods for details.

    This (including the documentation) is a partial translation of Guava's RangeSet to Python.
    Guava's implementation was written by Kevin Bourrillion and Louis Wasserman.
    """

    __slots__ = ()

    @staticmethod
    def create_mutable() -> "MutableRangeSet[T]":
        return _MutableSortedDictRangeSet.create()

    @abstractmethod
    def __contains__(self, value: T) -> bool:  # type: ignore
        """
        Determine whether any of this range set's member ranges contains `value`.
        """
        raise NotImplementedError()

    @abstractmethod
    def encloses(self, rng: Range[T]) -> bool:
        """
        Check if any member range encloses `rng`
        """
        raise NotImplementedError()

    def encloses_all(self, rngs: Union["RangeSet[T]", Iterable[Range[T]]]) -> bool:
        """
        For each input range, check if any member range encloses it.
        """
        if isinstance(rngs, RangeSet):
            return self.encloses_all(rngs.as_ranges())
        for rng in rngs:
            if not self.encloses(rng):
                return False
        return True

    @abstractmethod
    def intersects(self, rng: Range[T]) -> bool:
        """
        Get whether any ranges in this set intersects `rng`

        Returns `True` if there exists a non-empty range enclosed by both a member range in
        this range set and the specified range.
        """
        raise NotImplementedError()

    @abstractmethod
    def ranges_overlapping(self, rng: Range[T]) -> ImmutableSet[Range[T]]:
        """
        Get all ranges in this set that overlap (have an intersection) with `rng`.

        Unlike Guava's `intersectRanges`, this does not truncate partially intersecting ranges to
        just the intersecting portion.
        """
        raise NotImplementedError()

    @abstractmethod
    def range_containing(self, value: T) -> Optional[Range[T]]:
        raise NotImplementedError()

    @abstractmethod
    def range_enclosing_range(self, value: Range[T]) -> Optional[Range[T]]:
        raise NotImplementedError()

    @abstractmethod
    def ranges_enclosed_by(self, rng) -> ImmutableSet[Range[T]]:
        raise NotImplementedError()

    @abstractmethod
    def rightmost_containing_or_below(self, upper_limit: T) -> Optional[Range[T]]:
        """
        Get the rightmost range in this set whose lower bound does not exceed *upper_limit*.

        Formally, this is the range `(x, y)` with minimal `y` such that `(upper_limit, +inf)`
        does not contain `(x, y)`.

        If there is no such set, `None` is returned.

        For example::

         range_set: RangeSet[int] = immutablerangeset([
             Range.open_closed(1, 10)
             Range.open(12, 15)
         ])

         // range_set: {(1, 10], (12, 15)}
         range_set.rightmost_containing_or_below(3)  // returns (1, 10]
         range_set.rightmost_containing_or_below(11) // returns (1, 10]
         range_set.rightmost_containing_or_below(12) // returns (1, 10]
         range_set.rightmost_containing_or_below(13) // returns (12, 15)
         range_set.rightmost_containing_or_below(15) // returns (12, 15)
         range_set.rightmost_containing_or_below(1)  // returns None
        """

    @abstractmethod
    def leftmost_containing_or_above(self, lower_limit: T) -> Optional[Range[T]]:
        """
        Get the leftmost range in this set whose upper bound is not below *lower_limit*.

        Formally, this is the range `(x, y)` with maximal `x` such that `(-inf, lower_limit)`
        does not contain `(x, y)`.

        If there is no such set, `None` is returned.

        For example::

         range_set: RangeSet[int] = immutablerangeset([
             Range.open(1, 10)
             Range.open_closed(12, 15)
         ])

         // range_set: {(1, 10), (12, 15]}
         range_set.leftmost_containing_or_above(1)  // returns (1, 10)
         range_set.leftmost_containing_or_above(3)  // returns (1, 10)
         range_set.leftmost_containing_or_above(10) // returns (12, 15]
         range_set.leftmost_containing_or_above(11) // returns (12, 15]
         range_set.leftmost_containing_or_above(12) // returns (12, 15]
         range_set.leftmost_containing_or_above(13) // returns (12, 15]
         range_set.leftmost_containing_or_above(15) // returns (12, 15]
         range_set.leftmost_containing_or_above(16) // returns None
        """

    def maximal_containing_or_below(self, upper_limit: T) -> Optional[Range[T]]:
        """
        Deprecated. Instead, use rightmost_containing_or_below().

        This may be removed in future versions.
        """
        warnings.warn(
            "Deprecated, use rightmost_containing_or_below(upper_limit). "
            "This method may be removed in a future release.",
            DeprecationWarning,
        )
        return self.rightmost_containing_or_below(upper_limit)

    def minimal_containing_or_above(self, lower_limit: T) -> Optional[Range[T]]:
        """
        Deprecated. Instead, use leftmost_containing_or_below().

        This may be removed in future versions.
        """
        warnings.warn(
            "Deprecated, use leftmost_containing_or_above(upper_limit). "
            "This method may be removed in a future release.",
            DeprecationWarning,
        )
        return self.leftmost_containing_or_above(lower_limit)

    @abstractmethod
    def as_ranges(self) -> Sequence[Range[T]]:
        raise NotImplementedError()

    def __eq__(self, other) -> bool:
        if isinstance(other, RangeSet):
            return ImmutableSet.of(self.as_ranges()) == ImmutableSet.of(other.as_ranges())
        else:
            return False

    def __hash__(self) -> int:
        return hash(self.as_ranges())

    @abstractmethod
    def is_empty(self) -> bool:
        """
        Determine if this range set has no ranges.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def span(self) -> Range[T]:
        """
        The minimal range which encloses all ranges in this range set.
        """
        raise NotImplementedError()

    def __repr__(self):
        return self.__class__.__name__ + "(" + str(self.as_ranges()) + ")"

    def __getstate__(self) -> Tuple[Range[T], ...]:
        if self.is_empty():
            return ()
        return tuple(self.as_ranges())

    @abstractmethod
    def __setstate__(self, state: Iterable[Range[T]]) -> None:
        raise NotImplementedError()


T2 = TypeVar("T2")


class ImmutableRangeSet(RangeSet[T], metaclass=ABCMeta):
    """
    A RangeSet which cannot be modified.

    If you reach into its guts and modify it, its behavior is undefined.
    """

    @staticmethod
    def builder() -> "ImmutableRangeSet.Builder[T]":
        return _ImmutableSortedDictRangeSet.Builder()

    class Builder(Generic[T2], metaclass=ABCMeta):
        @abstractmethod
        def add(self, rng: Range[T2]) -> "ImmutableRangeSet.Builder[T2]":
            raise NotImplementedError()

        def add_all(self, ranges: Iterable[Range[T2]]) -> "ImmutableRangeSet.Builder[T2]":
            for rng in ranges:
                self.add(rng)
            return self

        @abstractmethod
        def build(self) -> "ImmutableRangeSet[T2]":
            raise NotImplementedError()


class MutableRangeSet(RangeSet[T], metaclass=ABCMeta):
    __slots__ = ()

    def add(self, rng: Range[T]) -> "MutableRangeSet[T]":
        """
        Add the specified range to this RangeSet (optional operation).

         For equal range sets `a` and `b`, the result of `a.add(range)` is that `a` will be the
         minimal range set for which both `a.encloses_all(b)` and `a.encloses(range)`.

        Note that `range` will be coalesced with any ranges in the range set that are connected
         with it. Moreover, if range is empty, this is a no-op.

        Returns the RangeSet itself to facilitate chaining operations, especially in tests.
         """
        raise NotImplementedError()

    def add_all(
        self, rngs: Union["RangeSet[T]", Iterable[Range[T]]]
    ) -> "MutableRangeSet[T]":
        """
        Add all the specified ranges to this RangeSet (optional operation).

        Returns the RangeSet itself to facilitate chaining operations, especially in tests.
        """
        if isinstance(rngs, RangeSet):
            return self.add_all(rngs.as_ranges())
        for rng in rngs:
            self.add(rng)
        return self

    def clear(self) -> None:
        """
        Remove all ranges from this RangeSet (optional operation).

        After this operation, `c in this_range_set` will return `False` for all `c`.

        This is equivalent to `remove(Range.all())`.
        """
        raise NotImplementedError()

    def remove(self, rng: Range[T]) -> "MutableRangeSet[T]":
        """
        Remove the specified range from this RangeSet (optional operation).

        After this operation, if `rng.contains(c)`, `self.contains(c)` will return `False`.
        If `rng` is empty, this is a no-op.

        Returns the RangeSet itself to facilitate chaining operations, especially in tests.
        """
        raise NotImplementedError()

    def remove_all(
        self, rngs: Union["RangeSet[T]", Iterable[Range[T]]]
    ) -> "MutableRangeSet[T]":
        """
        Remove each specified range.

        Returns the RangeSet itself to facilitate chaining operations, especially in tests.
        """
        if isinstance(rngs, RangeSet):
            return self.remove_all(rngs.as_ranges())
        for rng in rngs:
            self.remove(rng)
        return self


# noinspection PyProtectedMember
class _SortedDictRangeSet(RangeSet[T], metaclass=ABCMeta):
    # pylint:disable=protected-access

    def __init__(self, ranges_by_lower_bound: SortedDict) -> None:
        # we store the ranges as a map sorted by their lower bound
        # Note that because we enforce that there are no overlapping or connected ranges,
        # this sorts the ranges by upper bound as well
        self._ranges_by_lower_bound = ranges_by_lower_bound

    def range_containing(self, value: T) -> Optional[Range[T]]:
        highest_range_beginning_at_or_below = _value_at_or_below(
            self._ranges_by_lower_bound, _BelowValue(value)
        )
        if highest_range_beginning_at_or_below:
            if value in highest_range_beginning_at_or_below:
                return highest_range_beginning_at_or_below
        return None

    def range_enclosing_range(self, rng: Range[T]) -> Optional[Range[T]]:
        # this implementation can be sped up
        highest_range_beginning_at_or_below = _value_at_or_below(
            self._ranges_by_lower_bound, rng._lower_bound
        )
        if (
            highest_range_beginning_at_or_below
            and highest_range_beginning_at_or_below.encloses(rng)
        ):
            return highest_range_beginning_at_or_below
        return None

    def ranges_enclosed_by(self, query_rng) -> ImmutableSet[Range[T]]:
        highest_range_at_or_above = _value_at_or_above(
            self._ranges_by_lower_bound, query_rng._lower_bound
        )
        if highest_range_at_or_above:
            start_idx = self._ranges_by_lower_bound.values().index(
                highest_range_at_or_above
            )
            ret: ImmutableSet.Builder[Range[T]] = ImmutableSet.builder()
            for idx in range(start_idx, len(self._ranges_by_lower_bound)):
                rng_at_idx = self._ranges_by_lower_bound.values()[idx]
                if query_rng.encloses(rng_at_idx):
                    ret.add(rng_at_idx)
                else:
                    break
            return ret.build()
        else:
            return immutableset()

    # noinspection PyTypeHints
    def __contains__(self, value: T) -> bool:  # type: ignore
        highest_range_beginning_at_or_below = _value_at_or_below(
            self._ranges_by_lower_bound, _BelowValue(value)
        )
        return bool(
            highest_range_beginning_at_or_below
            and value in highest_range_beginning_at_or_below
        )

    def encloses(self, rng: Range[T]) -> bool:
        highest_range_beginning_at_or_below = _value_at_or_below(
            self._ranges_by_lower_bound, rng._lower_bound
        )
        return bool(
            highest_range_beginning_at_or_below
            and highest_range_beginning_at_or_below.encloses(rng)
        )

    def intersects(self, rng: Range[T]) -> bool:
        check_not_none(rng)
        ceiling_range: Optional[Range[T]] = _value_at_or_above(
            self._ranges_by_lower_bound, rng._lower_bound
        )
        if (
            ceiling_range
            and ceiling_range.is_connected(rng)
            and not ceiling_range.intersection(rng).is_empty()
        ):
            return True
        # check strictness of lowerEntry
        lower_range: Optional[Range[T]] = _value_below(
            self._ranges_by_lower_bound, rng._lower_bound
        )
        return bool(
            lower_range
            and lower_range.is_connected(rng)
            and not lower_range.intersection(rng).is_empty()
        )

    def ranges_overlapping(self, rng: Range[T]) -> ImmutableSet[Range[T]]:
        check_not_none(rng)
        if self.is_empty():
            return immutableset()
        rlb = self._ranges_by_lower_bound
        from_index = rlb.bisect(rng._lower_bound)
        # If we would insert at the end (are greater than all the elements, the only range that
        # could possibly overlap is the last one.
        if from_index == len(rlb):
            last_range: Range[T] = rlb[rlb.keys()[-1]]
            if last_range.intersects(rng):
                return immutableset([last_range])
            return immutableset()
        to_index = rlb.bisect(rng._upper_bound)
        # If we would insert at the start (are smaller than all the elements, the only range that
        # could possibly overlap is the first one.
        if to_index == 0:
            first_range: Range[T] = rlb[rlb.keys()[0]]
            if first_range.intersects(rng):
                return immutableset([first_range])
            return immutableset()
        return immutableset(
            [
                rlb[rlb.keys()[index]]
                # The ranges at the extreme indices do not necessarily overlap,
                for index in range(max(0, from_index - 1), min(to_index, len(rlb) - 1))
                # so this explicit check is needed.
                if rlb[rlb.keys()[index]].intersects(rng)
            ]
        )

    def rightmost_containing_or_below(self, upper_limit: T) -> Optional[Range[T]]:
        return _value_at_or_below(self._ranges_by_lower_bound, _BelowValue(upper_limit))

    def leftmost_containing_or_above(self, lower_limit: T) -> Optional[Range[T]]:
        sorted_dict = self._ranges_by_lower_bound
        # an AboveValue cut corresponds to a closed upper interval, which catches containment
        # as desired
        # I have no idea why mypy is asking for an explicit type assignment here
        limit_as_bound: _AboveValue = _AboveValue(lower_limit)

        # insertion index into the sorted list of sets
        idx = sorted_dict.bisect_left(limit_as_bound)
        # so the index of the "latest" set with a lower bound preceding lower_limit is back one
        containing_or_below_index = idx - 1

        if containing_or_below_index >= 0:
            # if such a set exists, we need to check if we are contained in it...
            latest_beginning_before = sorted_dict[
                sorted_dict.keys()[containing_or_below_index]
            ]
            if limit_as_bound <= latest_beginning_before._upper_bound:
                return latest_beginning_before

        if idx < len(sorted_dict):
            return sorted_dict[sorted_dict.keys()[idx]]
        else:
            return None

    def as_ranges(self) -> Sequence[Range[T]]:
        return self._ranges_by_lower_bound.values()

    def is_empty(self) -> bool:
        return len(self._ranges_by_lower_bound) == 0

    @property
    def span(self) -> Range[T]:
        if self.is_empty():
            raise ValueError("Can't take span of an empty RangeSet")
        return Range(
            self._ranges_by_lower_bound.values()[0]._lower_bound,
            self._ranges_by_lower_bound.values()[-1]._upper_bound,
        )

    def immutable_copy(self) -> ImmutableRangeSet[T]:
        return _ImmutableSortedDictRangeSet(self._ranges_by_lower_bound.copy())

    def __repr__(self):
        return repr(list(self.as_ranges()))

    def __len__(self) -> int:
        return len(self._ranges_by_lower_bound)

    def __setstate__(self, state: Iterable[Range[T]]) -> None:
        self._ranges_by_lower_bound = SortedDict(
            [(rng._lower_bound, rng) for rng in state]
        )


class _MutableSortedDictRangeSet(_SortedDictRangeSet[T], MutableRangeSet[T]):
    # pylint:disable=protected-access

    @staticmethod
    def create() -> "MutableRangeSet[T]":
        return _MutableSortedDictRangeSet(SortedDict())

    def add(self, range_to_add: Range[T]) -> "MutableRangeSet[T]":
        if range_to_add.is_empty():
            return self

        # the range we actually need to add may not correspond exactly to range_to_add
        # because it may overlap or be connected to existing ranges
        # lb_to_add and ub_to_add will form the range we actually add
        lb_to_add = range_to_add._lower_bound
        ub_to_add = range_to_add._upper_bound

        range_below_lb: Optional[Range[T]] = _value_below(
            self._ranges_by_lower_bound, lb_to_add
        )

        # is there any range which begins strictly before range_to_add's lower bound?
        # if so, range_to_add's beginning might lie within that range...
        if range_below_lb and lb_to_add <= range_below_lb._upper_bound:
            # and we need to coalesce with it by extending the bounds to include
            # that range's lower bound
            lb_to_add = range_below_lb._lower_bound
            # if the upper bound exceeds ours, too, then our range to add is entirely
            # contained within range_below_lb. Since this is already in the set, we
            # have nothing to do
            if ub_to_add < range_below_lb._upper_bound:
                return self

        # now we need to check of coalescing and connectedness on the upper end
        range_below_ub: Optional[Range[T]] = _value_at_or_below(
            self._ranges_by_lower_bound, ub_to_add
        )
        if range_below_ub and ub_to_add < range_below_ub._upper_bound:
            ub_to_add = range_below_ub._upper_bound

        # any ranges which lie within the range we are getting ready to add are subsumed in it
        _clear(self._ranges_by_lower_bound, lb_to_add, ub_to_add)

        self._replace_range_with_same_lower_bound(Range(lb_to_add, ub_to_add))
        return self

    def add_all(
        self, ranges_to_add: Union["RangeSet[T]", Iterable[Range[T]]]
    ) -> "MutableRangeSet[T]":
        if isinstance(ranges_to_add, RangeSet):
            return self.add_all(ranges_to_add.as_ranges())
        for rng in ranges_to_add:
            self.add(rng)
        return self

    def clear(self) -> None:
        _clear(self._ranges_by_lower_bound, _BELOW_ALL, _ABOVE_ALL)

    def remove(self, rng: Range[T]) -> "MutableRangeSet[T]":
        raise NotImplementedError(
            "I didn't need this, so I didn't bother to implement it yet."
        )

    def _replace_range_with_same_lower_bound(self, rng: Range[T]) -> None:
        if rng.is_empty():
            del self._ranges_by_lower_bound[rng._lower_bound]
        else:
            self._ranges_by_lower_bound[rng._lower_bound] = rng


class _ImmutableSortedDictRangeSet(ImmutableRangeSet[T], _SortedDictRangeSet[T]):
    """
    An implementation of ImmutableRangeSet

    This should never be directly created by the user. In particular if something maintains
    a reference to sorted_dict, then all immutability guarantees are broken.
    """

    class Builder(ImmutableRangeSet.Builder[T2]):
        def __init__(self):
            self._mutable_builder = _MutableSortedDictRangeSet.create()

        def add(self, rng: Range[T2]) -> "ImmutableRangeSet.Builder[T2]":
            self._mutable_builder.add(rng)
            return self

        def build(self) -> "ImmutableRangeSet[T2]":
            return self._mutable_builder.immutable_copy()


K = TypeVar("K")
V = TypeVar("V")


class RangeMap(Generic[K, V], metaclass=ABCMeta):
    """
    A mapping from disjoint nonempty ranges to values.

    Queries look up the value associated with the range (if any) that contains a specified key.

    In contrast to RangeSet, no "coalescing" is done of connected ranges, even if they are mapped
    to the same value.

    Note that this does not extend `Mapping` because you can't iterate over the keys.

    This (including the documentation) is a partial translation of Guava's RangeMap to Python.
    Guava's implementation was written by Louis Wasserman.
    """

    __slots__ = ()

    @abstractmethod
    def __contains__(self, key: K) -> bool:
        """
        Determine whether any of this range set's key ranges contains `key`.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_enclosed_by(self, rng: Range[K]) -> ImmutableSet[V]:
        """
        Get values mapped to by any key in `rng`.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_from_rightmost_containing_or_below(self, key: K):
        """
        Get the value associated with the rightmost range in this set whose lower bound does not
        exceed *upper_limit*.

        Formally, this is the value associated with the range `(x, y)` with minimal `y` such that
        `(upper_limit, +inf)` does not contain `(x, y)`.

        If there is no such set, `None` is returned.

        For example::

         range_map: RangeMap[int, int] = immutablerangemap([
           (Range.open_closed(1, 10), 36),  // (1, 10) maps to 36
           (Range.open(12, 15), 17)         // (12, 13) maps to 17
         ]);`

         // range keys: {(1, 10], (12, 15)}
         range_map.get_from_rightmost_containing_or_below(1)  // returns None
         range_map.get_from_rightmost_containing_or_below(3)  // returns 36
         range_map.get_from_rightmost_containing_or_below(11) // returns 36
         range_map.get_from_rightmost_containing_or_below(12) // returns 36
         range_map.get_from_rightmost_containing_or_below(13) // returns 17
         range_map.get_from_rightmost_containing_or_below(15) // returns 17
        """

    @abstractmethod
    def get_from_leftmost_containing_or_above(self, key: K):
        """
        Get the value associated with the leftmost range in this set whose upper bound is not below
        *lower_limit*.

        Formally, this is the value associated with the range `(x, y)` with maximal `x` such that
        `(-inf, lower_limit)` does not contain `(x, y)`.

        If there is no such set, `None` is returned.

        For example::

         range_map: RangeSet[int] = immutablerangemap([
             (Range.open(1, 10), 5),
             (Range.open_closed(12, 15), 7)
         ])

         // range keys: {(1, 10), (12, 15]}
         range_map.get_from_leftmost_containing_or_above(1)  // returns 5
         range_map.get_from_leftmost_containing_or_above(3)  // returns 5
         range_map.get_from_leftmost_containing_or_above(10) // returns 7
         range_map.get_from_leftmost_containing_or_above(11) // returns 7
         range_map.get_from_leftmost_containing_or_above(12) // returns 7
         range_map.get_from_leftmost_containing_or_above(13) // returns 7
         range_map.get_from_leftmost_containing_or_above(15) // returns 7
         range_map.get_from_leftmost_containing_or_above(16) // returns None
        """

    def get_from_maximal_containing_or_below(self, key: K):
        """
        Deprecated. Instead, use get_from_rightmost_containing_or_below(key).

        This may be removed in future versions.
        """
        warnings.warn(
            "Deprecated, use get_from_rightmost_containing_or_below(upper_limit). "
            "This method may be removed in a future release.",
            DeprecationWarning,
        )
        return self.get_from_rightmost_containing_or_below(key)

    def get_from_minimal_containing_or_above(self, key: K):
        """
        Deprecated. Instead, use get_from_leftmost_containing_or_above(key).

        This may be removed in future versions.
        """
        warnings.warn(
            "Deprecated, use get_from_leftmost_containing_or_below(key). "
            "This method may be removed in a future release.",
            DeprecationWarning,
        )
        return self.get_from_leftmost_containing_or_above(key)

    def __eq__(self, other) -> bool:
        if isinstance(other, RangeMap):
            return ImmutableSet.of(self.as_dict()) == ImmutableSet.of(other.as_dict())
        else:
            return False

    @abstractmethod
    def as_dict(self) -> Mapping[Range[K], V]:
        raise NotImplementedError()

    def __hash__(self) -> int:
        return hash(self.as_dict())

    @abstractmethod
    def is_empty(self) -> bool:
        """
        Determine if this range map has no mappings.
        """
        raise NotImplementedError()

    def __repr__(self):
        return self.__class__.__name__ + "(" + str(self.as_dict()) + ")"


# necessary for builder because an inner class cannot share type variables with its outer class
K2 = TypeVar("K2")
V2 = TypeVar("V2")


# this should have slots=True but cannot for the moment due to
# https://github.com/python-attrs/attrs/issues/313
@attrs(frozen=True, repr=False)
class ImmutableRangeMap(Generic[K, V], RangeMap[K, V]):
    rng_to_val: ImmutableDict[Range[K], V] = attrib(converter=immutabledict)
    range_set: ImmutableRangeSet[K] = attrib(init=False)

    def __attrs_post_init__(self) -> None:
        if len(self.rng_to_val) != len(self.range_set):
            raise ValueError(
                "Some range keys are connected or overlapping. Overlapping keys "
                "will never be supported. Support for connected keys is tracked in "
                "https://github.com/isi-vista/vistautils/issues/37"
            )

    @staticmethod
    def empty() -> "ImmutableRangeMap[K, V]":
        """Deprecated - prefer the module-level factory ``immutablerangemap`` with no arguments."""
        return ImmutableRangeMap(immutabledict())

    @staticmethod
    def builder() -> "ImmutableRangeMap.Builder[K, V]":
        return ImmutableRangeMap.Builder()

    def __contains__(self, key: K) -> bool:
        return key in self.range_set

    def get_enclosed_by(self, rng: Range[K]) -> ImmutableSet[V]:
        ret: ImmutableSet.Builder[V] = ImmutableSet.builder()
        for rng_key in self.range_set.ranges_enclosed_by(rng):
            ret.add(self.rng_to_val[rng_key])
        return ret.build()

    def is_empty(self) -> bool:
        return self.range_set.is_empty()

    def __getitem__(self, k: K) -> Optional[V]:
        containing_range = self.range_set.range_containing(k)
        return self.rng_to_val[containing_range] if containing_range else None

    def as_dict(self) -> Mapping[Range[K], V]:
        return self.rng_to_val

    def get_from_rightmost_containing_or_below(self, key: K) -> Optional[V]:
        probe_range = self.range_set.rightmost_containing_or_below(key)
        return self.rng_to_val[probe_range] if probe_range else None

    def get_from_leftmost_containing_or_above(self, key: K) -> Optional[V]:
        probe_range = self.range_set.leftmost_containing_or_above(key)
        return self.rng_to_val[probe_range] if probe_range else None

    def __reduce__(self):
        # __getstate__/__setstate__ cannot be used because the implementation is frozen.
        _repr = ()
        if not self.is_empty():
            _repr = tuple(self.as_dict().items())
        return (immutablerangemap, (_repr,))

    @range_set.default  # type: ignore
    def _init_range_set(self) -> ImmutableRangeSet[K]:
        return (  # type: ignore
            ImmutableRangeSet.builder()  # type: ignore
            .add_all(self.rng_to_val.keys())  # type: ignore
            .build()  # type: ignore
        )

    class Builder(Generic[K2, V2]):
        def __init__(self):
            self.rng_to_val = ImmutableDict.builder()

        def put(self, key: Range[K2], val: V2) -> "ImmutableRangeMap.Builder[K2, V2]":
            self.rng_to_val.put(key, val)
            return self

        def build(self) -> "ImmutableRangeMap[K2, V2]":
            return ImmutableRangeMap(self.rng_to_val.build())


def immutablerangemap(
    mappings: Optional[Iterable[Tuple[Range[K], V]]] = None
) -> ImmutableRangeMap[K, V]:
    return ImmutableRangeMap(immutabledict(mappings))


# utility functions for SortedDict to give it an interface more like Java's NavigableMap
def _value_below(sorted_dict: SortedDict, key: T) -> Optional[Any]:
    """
    Get item for greatest key strictly less than the given key

    Returns None if there is no such key.
    """
    idx = sorted_dict.bisect_left(key) - 1
    if idx >= 0:
        if idx >= len(sorted_dict):
            idx = len(sorted_dict) - 1
        lb_key = sorted_dict.keys()[idx]
        return sorted_dict[lb_key]
    else:
        return None


def _value_at_or_below(sorted_dict: SortedDict, key: T) -> Optional[Any]:
    """
    Get item for greatest key less than or equal to a given key.

    Returns None if there is no such key
    """
    if not sorted_dict:
        return None

    idx = sorted_dict.bisect_left(key)

    if idx >= len(sorted_dict) or key != sorted_dict.keys()[idx]:
        if idx > 0:
            key = sorted_dict.keys()[idx - 1]
        else:
            return None
    return sorted_dict[key]


def _value_at_or_above(sorted_dict: SortedDict, key: T) -> Optional[Any]:
    if not sorted_dict:
        return None
    idx = sorted_dict.bisect_left(key)
    if idx >= len(sorted_dict):
        return None

    return sorted_dict[sorted_dict.keys()[idx]]


def _clear(
    sorted_dict: SortedDict, start_key_inclusive: T, stop_key_exclusive: T
) -> None:
    # we copy to a list first in case sorted_dict is not happy with modification during iteration
    for key_to_delete in list(
        sorted_dict.irange(
            start_key_inclusive, stop_key_exclusive, inclusive=(True, False)
        )
    ):
        del sorted_dict[key_to_delete]



# pylint: skip-file
import inspect
import logging
import os
import re
import shutil
from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import yaml
from attr import attrib, attrs, evolve
from immutablecollections import ImmutableDict, immutabledict
from immutablecollections.converter_utils import _to_tuple

from vistautils._graph import Digraph
from vistautils.io_utils import CharSink, is_empty_directory
from vistautils.misc_utils import eval_in_context_of_modules
from vistautils.preconditions import check_arg, check_isinstance
from vistautils.range import Range

_logger = logging.getLogger(__name__)  # pylint:disable=invalid-name


class ParameterError(Exception):
    pass


_ParamType = TypeVar("_ParamType")  # pylint:disable=invalid-name
_U = TypeVar("_U")  # pylint:disable=invalid-name


@attrs(frozen=True, slots=True)
class Parameters:
    """
    Configuration parameters for a program.

    A `Parameters` object can be thought of as a hierarchical dictionary mapping parameter name
    strings to arbitrary values.  Hierarchical levels within a parameter name can be indicated
    via `.`; the levels are called namespaces. So `hello.world.foo` means "look up the 'foo'
    parameter within the namespace 'world' which is itself within the namespace 'hello'.

    The advantage of using this class over simple dictionary of dictionaries is that it provides
    type-safe accessors for parameters which can do validation and ensure certain conditions hold.
    This tends to push the discovery of errors earlier in program execution (or even before program
    execution).

    `Parameters` objects can be represented in multiple formats. The only loader currently
    implemented is `YAMLParametersLoader`.

    You can check if a lookup of a parameter would be successful using the `in` operator.
    """

    _data: ImmutableDict[str, Any] = attrib(
        default=immutabledict(), converter=immutabledict
    )
    namespace_prefix: Tuple[str, ...] = attrib(
        default=tuple(), converter=_to_tuple, kw_only=True
    )

    def __attrs_post_init__(self) -> None:
        for key in self._data:
            check_arg(
                "." not in key, "Parameter keys cannot contain namespace separator '.'"
            )

    @staticmethod
    def empty(*, namespace_prefix: Iterable[str] = tuple()) -> "Parameters":
        """
        A `Parameters` with no parameter mappings.
        """
        return Parameters.from_mapping(
            ImmutableDict.empty(), namespace_prefix=namespace_prefix
        )

    @staticmethod
    def from_mapping(
        mapping: Mapping, *, namespace_prefix: Iterable[str] = tuple()
    ) -> "Parameters":
        """
        Convert a dictionary of dictionaries into a `Parameter`s

        The top-level dictionary becomes the top-level namespace.  Each mapping-valued parameter
        becomes a namespace.
        """
        check_isinstance(mapping, Mapping)
        ret: List[Tuple[str, Any]] = []
        for (key, val) in mapping.items():
            if isinstance(val, Mapping):
                sub_namespace_prefix = list(namespace_prefix)
                sub_namespace_prefix.append(key)
                ret.append(
                    (
                        key,
                        Parameters.from_mapping(
                            val, namespace_prefix=sub_namespace_prefix
                        ),
                    )
                )
            else:
                # this case will also be triggered if the value is already a parameters object
                ret.append((key, val))
        return Parameters(ret, namespace_prefix=namespace_prefix)

    def as_mapping(self) -> Mapping[str, Any]:
        """
        Get these parameter values as a ``Mapping``.
        """
        return self._data

    def creatable_directory(self, param: str) -> Path:
        """
        Get a directory which can be written to.

        Interprets the string-valued parameter `param` as a directory path and creates it if
        it does not exist.  This allows writing files within this directory without needing to
        remember to create it or its parent directories.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        ret.mkdir(parents=True, exist_ok=True)
        return ret

    def optional_creatable_directory(self, param: str) -> Optional[Path]:
        """
        Get a directory which can be written to, if possible.

        If *param* is not present, returns *None*.

        Interprets the string-valued parameter `param` as a directory path and creates it if
        it does not exist.  This allows writing files within this directory without needing to
        remember to create it or its parent directories.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        if param in self:
            return self.creatable_directory(param)
        else:
            return None

    def creatable_empty_directory(self, param: str, *, delete=False) -> Path:
        """
        Get an empty directory which can be written to.

        Interprets the string-valued parameter `param` as a directory path and creates it if
        it does not exist.  If the directory already exists and is non-empty, behavior depends
        on the value of the `delete` argument.  If False (the default), an exception will be
        raised.  If True, the directory and its contents will be deleted.

        This allows writing files within this directory without needing to remember to create it
        or its parent directories.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        if ret.is_dir():
            if not is_empty_directory(ret):
                if delete:
                    shutil.rmtree(str(ret))
                    ret.mkdir(parents=True, exist_ok=True)
                else:
                    raise ParameterError(
                        "Expected an empty directory for parameters {!s},"
                        "but got non-empty path {!s}".format(param, ret)
                    )
            return ret
        elif ret.exists():
            raise ParameterError(
                "Expected an empty directory for parameters {!s},"
                "but got non-directory {!s}".format(param, ret)
            )
        else:
            return self.creatable_directory(param)

    def optional_creatable_empty_directory(
        self, param: str, *, delete: bool = False
    ) -> Optional[Path]:
        """
        Get an empty directory which can be written to, if possible.

        If *param* is absent, returns *None*.

        Interprets the string-valued parameter `param` as a directory path and creates it if
        it does not exist.  If the directory already exists and is non-empty, behavior depends
        on the value of the `delete` argument.  If False (the default), an exception will be
        raised.  If True, the directory and its contents will be deleted.

        This allows writing files within this directory without needing to remember to create it
        or its parent directories.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        if param in self:
            return self.creatable_empty_directory(param, delete=delete)
        else:
            return None

    def creatable_file(self, param: str) -> Path:
        """
        Get a file path which can be written to.

        Interprets the string-valued parameter `param` as a file path and creates its parent
        directory if it does not exist.  This allows writing to this file without needing to
        remember to create its parent directories.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        ret.parent.mkdir(parents=True, exist_ok=True)
        return ret

    def optional_creatable_file(self, param: str) -> Optional[Path]:
        """
        Get a file path which can be written to, if specified.

        Just like `creatable_file` but returns `None` if the parameter is absent.
        """
        if param in self:
            return self.creatable_file(param)
        else:
            return None

    def existing_file(self, param: str) -> Path:
        """
        Gets a path for an existing file.

        Interprets the string-valued parameter `param` as a file path. Throws a `ParameterError`
        if the path does not exist or is not a file.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        if ret.exists():
            if ret.is_file():
                return ret
            else:
                raise ParameterError(
                    "For parameter "
                    + param
                    + ", expected an existing file but got existing non-file "
                    + str(ret)
                )
        else:
            raise ParameterError(
                "For parameter "
                + param
                + ", expected an existing file but got non-existent "
                + str(ret)
            )

    def optional_existing_file(self, param: str) -> Optional[Path]:
        """
        Gets a path for an existing file, if specified.

        Interprets the string-valued parameter `param` as a file path. Returns `None`
        if no parameter by that name is present.  Throws a `ParameterError`
        if the path does not exist.
        """
        if param in self:
            return self.existing_file(param)
        else:
            return None

    def existing_directory(self, param: str) -> Path:
        """
        Gets a path for an existing directory.

        Interprets the string-valued parameter `param` as a directory path. Throws a
        `ParameterError` if the path does not exist or is not a directory.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        if ret.exists():
            if ret.is_dir():
                return ret
            else:
                raise ParameterError(
                    "For parameter "
                    + param
                    + ", expected an existing directory but got existing "
                    "non-directory " + str(ret)
                )
        else:
            raise ParameterError(
                "For parameter "
                + param
                + ", expected an existing directory but got non-existent "
                + str(ret)
            )

    def optional_existing_directory(self, param: str) -> Optional[Path]:
        """
        Gets a path for an existing directory, if specified.

        Interprets the string-valued parameter `param` as a directory path.
        If the parameter is not present, returns `None`.  Throws a
        `ParameterError` if the path does not exist.
        """
        if param in self:
            return self.existing_directory(param)
        else:
            return None

    def string(
        self, param_name: str, valid_options: Optional[Iterable[str]] = None
    ) -> str:
        """
        Gets a string-valued parameter.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = self.get(param_name, str)
        if valid_options is not None and ret not in valid_options:
            raise ParameterError(
                f"The value {ret} for the parameter {param_name} is not one of the valid options "
                f"{tuple(valid_options)}"
            )
        return ret

    @overload
    def optional_string(
        self, param_name: str, valid_options: Optional[Iterable[str]] = None
    ) -> Optional[str]:
        ...

    @overload
    def optional_string(
        self,
        param_name: str,
        valid_options: Optional[Iterable[str]] = None,
        *,
        default: str,
    ) -> str:
        ...

    def optional_string(
        self,
        param_name: str,
        valid_options: Optional[Iterable[str]] = None,
        *,
        default: Optional[str] = None,
    ):
        """
        Gets a string-valued parameter, if possible.
        If a default is provided, return the default
        else returns *None* if the parameter is absent.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        if param_name in self:
            return self.string(param_name, valid_options)
        else:
            return default  # type: ignore

    def __contains__(self, param_name: str) -> bool:
        return self._private_get(param_name, optional=True) is not None

    def namespace(self, name: str) -> "Parameters":
        """
        Get the namespace with the given name.
        """
        return self.get(name, Parameters)

    def has_namespace(self, name: str) -> bool:
        """
        Returns whether the parameter of the specified *name* has a value
        which is a nested `Parameters`.
        """
        maybe_namespace = self._private_get(name, optional=True)
        return maybe_namespace is not None and isinstance(maybe_namespace, Parameters)

    def integer(self, name: str) -> int:
        """
        Gets an integer parameter.
        """
        return self.get(name, int)

    @overload
    def optional_integer(self, name: str) -> Optional[int]:
        ...

    @overload
    def optional_integer(self, name, *, default: int) -> int:
        ...

    def optional_integer(self, name: str, *, default: Optional[int] = None):
        """
        Gets an integer parameter, if possible.

        Returns *None* if the parameter is not present.
        """
        if name in self:
            return self.integer(name)
        else:
            return default  # type: ignore

    def positive_integer(self, name: str) -> int:
        """
        Gets an parameter with a positive integer value.

        Throws an exception if the parameter is present but is not a positive integer.
        """
        ret = self.integer(name)
        if ret > 0:
            return ret
        else:
            raise ParameterError(
                "For parameter {!s}, expected a positive integer but got {!s}".format(
                    name, ret
                )
            )

    @overload
    def optional_positive_integer(self, name: str) -> Optional[int]:
        ...

    @overload
    def optional_positive_integer(self, name: str, *, default: int) -> int:
        ...

    def optional_positive_integer(self, name: str, *, default: Optional[int] = None):
        """
        Gets a positive integer parameter, if possible.

        Returns *None* if the parameter is not present.
        Throws an exception if the parameter is present but is not a positive integer.
        """
        if name in self:
            return self.positive_integer(name)
        if default:
            if isinstance(default, int) and default > 0:
                return default  # type: ignore
            else:
                raise ParameterError(f"Default value: {default} is not a positive value")
        return None

    def floating_point(
        self, name: str, valid_range: Optional[Range[float]] = None
    ) -> float:
        """
        Gets a float parameter.

        Throws a `ParameterError` if `param` is not within the given range.

        This method isn't called `float` to avoid a clash with the Python type.
        """
        ret = self.get(name, float)
        if valid_range is not None and ret not in valid_range:
            raise ParameterError(
                "For parameter {!s}, expected a float in the range {!s} but got {!s}".format(
                    name, valid_range, ret
                )
            )
        return ret

    @overload
    def optional_floating_point(
        self, name: str, valid_range: Optional[Range[float]] = None
    ) -> Optional[float]:
        ...

    @overload
    def optional_floating_point(
        self, name: str, valid_range: Optional[Range[float]] = None, *, default: float
    ) -> float:
        ...

    def optional_floating_point(
        self,
        name: str,
        valid_range: Optional[Range[float]] = None,
        *,
        default: Optional[float] = None,
    ):
        """
        Gets a float parameter if present.

        Throws a `ParameterError` if `param` is not within the given range.
        """
        if name in self:
            return self.floating_point(name, valid_range)
        if default:
            if (
                valid_range is not None
                and isinstance(default, float)
                and default not in valid_range
            ):
                raise ParameterError(
                    f"Default value of {default} not in the range of {valid_range}."
                )
            return default  # type: ignore
        else:
            return None

    def optional_float(
        self, name: str, valid_range: Optional[Range[float]] = None
    ) -> Optional[float]:
        """
        Deprecated, prefer `optional_floating_point` for more consistent naming.
        """
        return self.optional_floating_point(name, valid_range)

    def boolean(self, name: str) -> bool:
        """
        Gets a boolean parameter.
        """
        return self.get(name, bool)

    @overload
    def optional_boolean(self, name: str) -> Optional[bool]:
        ...

    @overload
    def optional_boolean(self, name: str, *, default: bool) -> bool:
        ...

    def optional_boolean(self, name: str, *, default: Optional[bool] = None):
        """
        Gets a boolean parameter if present.

        Avoid the temptation to do `params.optional_boolean('foo') or default_value`.
        """
        return self.get_optional(name, bool, default=default)

    def optional_boolean_with_default(self, name: str, default_value: bool) -> bool:
        """
        Deprecated. Prefer `optional_boolean` with default as a parameter.

        Gets a boolean parameter if present; otherwise returns the provided default.
        """
        ret = self.optional_boolean(name, default=default_value)
        if ret is not None:
            return ret
        else:
            return default_value

    def optional_namespace(self, name: str) -> Optional["Parameters"]:
        """
        Get the namespace with the given name, if possible.

        If the namespace is not present or there is a non-namespace parameter of the given name,
        If the namespace is not present or there is a non-namespace parameter of the given name,
        `None` is returned.
        """
        ret = self.get_optional(name, object)
        if isinstance(ret, Parameters):
            return ret
        else:
            return None

    def namespace_or_empty(self, name: str) -> Optional["Parameters"]:
        """
        Get the namespace with the given name, or an empty one.

        If the namespace is present, return it.
        If the parameter is present but is not a namespace, throw an exception.
        If the namespace is absent, return an empty namespace with the appropriate path prefix.
        """
        ret = self.get_optional(name, object)
        if isinstance(ret, Parameters):
            return ret
        elif ret is None:
            sub_namespace_prefix = list(self.namespace_prefix)
            sub_namespace_prefix.append(name)
            return Parameters.empty(namespace_prefix=sub_namespace_prefix)
        else:
            raise ParameterError(
                f"Expected a namespace, but got a regular parameters for {name}"
            )

    def arbitrary_list(self, name: str) -> List:
        """
        Get a list with arbitrary structure.
        """
        return self.get(name, List)

    @overload
    def optional_arbitrary_list(self, name: str) -> Optional[List]:
        ...

    @overload
    def optional_arbitrary_list(self, name: str, *, default: List) -> List:
        ...

    def optional_arbitrary_list(self, name: str, *, default: Optional[List] = None):
        """
        Get a list with arbitrary structure, if available
        """
        if not default:
            return self.get_optional(name, List)
        elif isinstance(default, List):
            return self.get_optional(name, List, default=default)

        raise ParameterError(
            f"Provided default to optional arbitrary list isn't a list. {default}"
        )

    def optional_evaluate(
        self,
        name: str,
        expected_type: Type[_ParamType],
        *,
        namespace_param_name: str = "value",
        special_values: Mapping[str, str] = ImmutableDict.empty(),
    ) -> Optional[_ParamType]:
        """
        Get a parameter, if present, interpreting its value as Python code.

        Same as `evaluate` except returning `None` if the requested parameter is not present.
        """
        if name in self:
            return self.evaluate(
                name,
                expected_type,
                namespace_param_name=namespace_param_name,
                special_values=special_values,
            )
        else:
            return None

    # type ignored because ImmutableDict.empty() has type Dict[Any, Any]
    def evaluate(
        self,
        name: str,
        expected_type: Type[_ParamType],
        *,
        context: Optional[Mapping] = None,
        namespace_param_name: str = "value",
        special_values: Mapping[str, str] = ImmutableDict.empty(),
    ) -> _ParamType:
        """
        Get a parameter, interpreting its value as Python code.

        The result of the code evaluation will be the parameter value. In addition to being a
        simple parameter, "name" can point to a namespace. In this case, the "value" parameter
        within the namespace is evaluated (unless some other parameter has been specified as
        `namespace_param_name`), but the list of modules given by the "import" parameter
        in that namespace (if present) are first imported.

        If the type of the result of the evaluation doesn't match `expected_type`, a
        `ParameterError` is raised.

        If `context` is specified, evaluation will happen in the context given. If you want
        evaluation to happen in the calling context, pass `locals()`.
        If the namespace contains the parameter *import*, it will be interpreted
        as a list of modules to import into the context before evaluation.

        Sometimes is it convenient to provide shortcuts for common cases. These can be specified
        in a `special_values` map whose keys are the special case values and whose values are the
        strings of expressions to be evaluated.
        """

        def handle_special_values(val: str) -> str:
            return special_values.get(val, val)

        namespace = self.optional_namespace(name)
        try:
            if namespace:
                return eval_in_context_of_modules(
                    handle_special_values(namespace.string(namespace_param_name)),
                    context or locals(),
                    context_modules=namespace.optional_arbitrary_list("import") or [],
                    expected_type=expected_type,
                )
            else:
                return eval_in_context_of_modules(
                    handle_special_values(self.string(name)),
                    context or locals(),
                    context_modules=[],
                    expected_type=expected_type,
                )
        except Exception as e:
            raise ParameterError(
                "Error while evaluating parameter {!s}".format(name)
            ) from e

    # type ignored because ImmutableDict.empty() has type Dict[Any, Any]
    def object_from_parameters(
        self,
        name: str,
        expected_type: Type[_ParamType],
        *,
        context: Optional[Mapping] = None,
        creator_namepace_param_name: str = "value",
        special_creator_values: Mapping[str, str] = ImmutableDict.empty(),
        default_creator: Optional[Any] = None,
    ) -> _ParamType:
        """
        Get an object of `expected_type`, initialized by the parameters in `name`.

        If `name` is a namespace, the `value` parameter within it is evaluated to get a "creator".
        If the result of the evaluation is a function, that function is itself the creator. If
        the result of the evaluation is a class, its `from_parameters` static method
        taking a single `Parameters` object will be used as the creator, if it exists. Otherwise
        its constructor will be used without parameters. The creator is then
        called with the namespace as its argument and the result is returned.  If the result does
        not match `expected_type` an exception will be raised. Do not include generic type arguments
        in `expected_type`.

        If `name` is a string, the same process is followed exception the string is evaluated
        directly to get the "creator" and an empty `Parameters` is passed.

        You can specify a different field within a namespace to be evaluated besides 'value' by
        specifying `creator_namespace_param_name`.

        You can specify a default creator to be used if none is specified with `default_creator`.

        You may specify additional context within which evaluation should happen with `context`.
        If you want evaluation to happen in the calling context, set this to `locals()`.
        If the namespace contains the parameter *import*, it will be interpreted
        as a list of modules to import into the context before evaluation.

        For the user's convenience, you may include a map of special values to expression strings.
        If the expression to be evaluated exactly matches any key of this map, the value from the
        map will be evaluated instead.

        If the namespace contains the field `import`, it will be treated as a comma-separated list
        of modules to be imported before evaluation.
        """
        if name in self:
            creator = self.evaluate(
                name,
                object,
                context=context,
                namespace_param_name=creator_namepace_param_name,
                special_values=special_creator_values,
            )
            if self.has_namespace(name):
                params_to_pass = self.namespace(name)
            else:
                params_to_pass = Parameters.empty(
                    namespace_prefix=_extend_prefix(self.namespace_prefix, name)
                )
        elif default_creator:
            creator = default_creator
            params_to_pass = Parameters.empty(
                namespace_prefix=_extend_prefix(self.namespace_prefix, name)
            )
        else:
            raise ParameterError(
                "No creator class specified when creating an object from {!s}".format(
                    name
                )
            )

        if inspect.isclass(creator):
            if hasattr(creator, "from_parameters"):
                ret: Callable[[Optional[Parameters]], _ParamType] = getattr(
                    creator, "from_parameters"
                )(params_to_pass)
            else:
                ret = creator()  # type: ignore
        elif callable(creator):
            ret = creator(params_to_pass)
        else:
            raise ParameterError(
                "Expected a class with from_parameters or a callable but got {!s}".format(
                    creator
                )
            )

        if isinstance(ret, expected_type):
            return ret
        else:
            raise ParameterError(
                "When instantiating using from_parameters, expected {!s} but"
                " got {!s}".format(expected_type, ret)
            )

    def get(self, param_name: str, param_type: Type[_ParamType]) -> _ParamType:
        """
        Get a parameter with type-safety.

        Gets the given parameter, throwing a `ParameterError` if it is not of the specified type.

        Throws a `ParameterError` if the parameter is unknown.
        """

        ret = self._private_get(param_name)
        if isinstance(ret, param_type):
            return ret
        else:
            raise ParameterError(
                f"{self._namespace_message()}When looking up parameter '{param_name}', "
                f"expected a value of type {param_type}, but got {ret} "
                "of type {type(ret)}"
            )

    @overload
    def get_optional(
        self, param_name: str, param_type: Type[_ParamType]
    ) -> Optional[_ParamType]:
        ...

    @overload
    def get_optional(
        self, param_name: str, param_type: Type[_ParamType], *, default: _U
    ) -> _U:
        ...

    def get_optional(
        self, param_name: str, param_type: Type[_ParamType], *, default: _U = None
    ):
        """
        Get a parameter with type-safety.

        Gets the given parameter, throwing a `ParameterError` if it is not of the
        specified type.

        If a default is provided return the default otherwise
        If the parameter is unknown, returns `None`
        """
        ret = self._private_get(param_name, optional=True)
        if ret is not None:
            if isinstance(ret, param_type):
                return ret
            else:
                raise ParameterError(
                    "When looking up parameter '{!s}', expected a value of type {!s}, but got {!s} "
                    "of type {!s}".format(param_name, param_type, ret, type(ret))
                )
        else:
            return default

    def path_list_from_file(self, param: str, *, log_name=None) -> Sequence[Path]:
        """
        Gets a list of paths from the file pointed to by param

        The paths are assumed to be listed one-per-line. Blank lines and lines
        where the first non-whitespace character is '#' are skipped.

        If log_name is specified, a message will be logged at info level of the form "Loaded
        <number> <log_name> from <file>"
        """
        file_list_file = self.existing_file(param)
        with open(str(file_list_file), "r", encoding="utf-8") as inp:
            ret = [
                Path(line.strip())
                for line in inp
                if line.strip() and not line.strip().startswith("#")
            ]
            if log_name:
                _logger.info("Loaded %s %s from %s", len(ret), log_name, file_list_file)
            return ret

    def path_map_from_file(self, param: str, *, log_name=None) -> Mapping[str, Path]:
        """
        Gets a map of keys to paths from the file pointed to by param

        We assume there are two tab-separated fields.  The first is the string key,
        the second is the path.

        If log_name is specified, a message will be logged at info level of the form "Loaded
        <number> <log_name> from <file>"
        """
        file_map_file = self.existing_file(param)
        with open(str(file_map_file), encoding="utf-8") as inp:
            ret_b: ImmutableDict.Builder[str, Path] = ImmutableDict.builder()
            for (line_num, line) in enumerate(inp):
                try:
                    parts = line.split("\t")
                    if len(parts) != 2:
                        raise IOError(
                            "Expected two tab-separated fields but got {!s}".format(
                                len(parts)
                            )
                        )
                    ret_b.put(parts[0].strip(), Path(parts[1].strip()))
                except Exception as e:
                    raise IOError(
                        "Error parsing line {!s} of {!s}:\n{!s}".format(
                            line_num, file_map_file, line
                        )
                    ) from e

            ret = ret_b.build()
            if log_name:
                _logger.info("Loaded %s %s from %s", len(ret), log_name, file_map_file)
            return ret

    def _private_get(self, param_name: str, optional: bool = False) -> Any:
        check_arg(isinstance(param_name, str))
        # pylint:disable=protected-access
        param_components = param_name.split(".")
        check_arg(param_components, "Parameter name cannot be empty")

        current = self
        namespaces_processed = []
        for param_component in param_components:
            if not isinstance(current, Parameters):
                if optional:
                    return None
                else:
                    raise ParameterError(
                        self._namespace_message()
                        + "When getting parameter "
                        + param_name
                        + " expected "
                        + ".".join(namespaces_processed)
                        + " to be a map, but it is a leaf: "
                        + str(current)
                        + ". Maybe you mistakenly prefixed the map keys with '-'?"
                    )

            if param_component in current._data:
                current = current._data[param_component]
                namespaces_processed.append(param_component)
            elif not optional:
                if namespaces_processed:
                    context_string = "in context " + ".".join(namespaces_processed)
                else:
                    context_string = "in root context"
                available_parameters = str(
                    [
                        key
                        for (key, val) in current._data.items()
                        if not isinstance(val, Parameters)
                    ]
                )
                available_namespaces = str(
                    [
                        key
                        for (key, val) in current._data.items()
                        if isinstance(val, Parameters)
                    ]
                )
                raise ParameterError(
                    self._namespace_message()
                    + "Parameter "
                    + param_name
                    + " not found. In "
                    + context_string
                    + " available parameters are "
                    + available_parameters
                    + ", available namespaces are "
                    + available_namespaces
                )
            else:
                # absent optional parameter
                return None

        return current

    def __str__(self) -> str:
        str_sink = CharSink.to_string()
        YAMLParametersWriter().write(self, str_sink)
        return str_sink.last_string_written

    def _namespace_message(self) -> str:
        if self.namespace_prefix:
            namespace_str = ".".join(self.namespace_prefix)
            return f"In namespace {namespace_str}: "
        else:
            return ""


def _extend_prefix(
    namespace_prefix: Tuple[str, ...], new_element: str
) -> Tuple[str, ...]:
    ret = list(namespace_prefix)
    ret.append(new_element)
    return tuple(new_element)


@attrs(auto_attribs=True)
class YAMLParametersLoader:
    """
    Loads `Parameters` from a modified YAML format.

    The format of the parameters file is YAML with the following restrictions:
    * all non-leaf objects, including the top-level, must be maps
    * all keys must be strings
    * duplicate keys are not allowed (this is not currently enforced: #259)

    The following additional processing will be done:
    * if there is a top-level list-valued key '_includes', it will be removed.  Each of its
    values will be interpreted as a path and loaded.  If there are multiple included files,
    parameters loaded from earlier files will form the loading context (See `load`) for later ones.
    Parameter mappings in files loaded later will overwrite those in files loaded earlier.  Relative
    paths will be interpreted relative to the parameter file being loaded.

    * any parameters containing strings surrounded by `%` will be interpolated by replacing that
    string (and the `%`s) with the value of the parameter.  Failed interpolations will result
    in a `ParameterError`.

    If *interpolate_environmental_variables* (default: true) is specified, then environmental
    variables will be available for interpolation, though they will not themselves appear in the
    loaded parameters.  Explicitly-specified parameters have priority over environmental variables.

    See unit tests in `test_parameters` for examples.
    """

    interpolate_environmental_variables: bool = True

    def load(
        self,
        f: Union[str, Path],
        context: Optional[Parameters] = None,
        *,
        included_context: Optional[Parameters] = None,
        namespace_path: Sequence[str] = tuple(),
    ):
        """
        Loads parameters from a YAML file.

        If `included_context` is specified, its content will be included in the returned
        Parameters (if not overridden) and will be available for interpolation.
        """

        # handle deprecated context parameter
        if context is not None:
            if included_context is not None:
                raise ParameterError(
                    "Cannot specify both included_context and deprecated context"
                    "parameters. Only specify the former."
                )
            else:
                included_context = context

        # provide default value for included context
        if included_context is None:
            non_none_included_context = Parameters.empty()
        else:
            non_none_included_context = included_context

        if isinstance(f, str):
            f = Path(f)

        return self._inner_load_from_string(
            f.read_text(encoding="utf-8"),
            error_string=str(f),
            includes_are_relative_to=f.parent,
            included_context=non_none_included_context,
            namespace_path=namespace_path,
        )

    def load_string(
        self,
        param_file_content: str,
        *,
        included_context: Parameters = Parameters.empty(),
        namespace_path: Sequence[str] = tuple(),
    ) -> Parameters:
        """
        Loads parameters from a string.

        This behaves just like *load*, except relative includes are not allowed.
        """
        return self._inner_load_from_string(
            param_file_content,
            error_string=f"String param file:\n{param_file_content}",
            includes_are_relative_to=None,
            included_context=included_context,
            namespace_path=namespace_path,
        )

    def _inner_load_from_string(
        self,
        param_file_content: str,
        error_string: str,
        *,
        included_context: Parameters = Parameters.empty(),
        includes_are_relative_to: Optional[Path] = None,
        namespace_path: Sequence[str] = tuple(),
    ):
        """
        Loads parameters from a YAML file.

        If `context` is specified, its content will be included in the returned Parameters (if
        not overridden) and will be available for interpolation.
        """
        try:
            raw_yaml = yaml.safe_load(param_file_content)
            self._validate(raw_yaml)
            previously_loaded = included_context

            # process and remove special include directives
            if "_includes" in raw_yaml:
                for included_file in raw_yaml["_includes"]:
                    _logger.info("Processing included parameter file %s", included_file)
                    if not os.path.isabs(included_file):
                        if includes_are_relative_to is not None:
                            included_file_path = Path(
                                includes_are_relative_to, included_file
                            ).resolve(strict=True)
                        else:
                            raise ParameterError(
                                "Cannot do relative includes when loading from"
                                "a string."
                            )
                    else:
                        included_file_path = Path(included_file)
                    previously_loaded = self._unify(
                        previously_loaded,
                        self._inner_load_from_string(
                            included_file_path.read_text(encoding="utf-8"),
                            error_string=str(included_file_path),
                            includes_are_relative_to=included_file_path.parent,
                            included_context=previously_loaded,
                        ),
                    )
                del raw_yaml["_includes"]

            interpolation_context = dict(previously_loaded.as_mapping())
            if self.interpolate_environmental_variables:
                for k, v in os.environ.items():
                    # environmental variables are overridden by explicit parameters
                    if k not in interpolation_context:
                        interpolation_context[k] = v

            return self._unify(
                previously_loaded,
                self._interpolate(
                    Parameters.from_mapping(raw_yaml),
                    Parameters.from_mapping(interpolation_context),
                ),
                namespace_prefix=namespace_path,
            )
        except Exception as e:
            raise IOError(f"Failure while loading parameter file {error_string}") from e

    @staticmethod
    def _validate(raw_yaml: Mapping):
        # we don't use check_isinstance so we can have a custom error message
        check_arg(
            isinstance(raw_yaml, Mapping),
            "Parameters YAML files must be mappings at the top level",
        )
        YAMLParametersLoader._check_all_keys_strings(raw_yaml)

    @staticmethod
    def _check_all_keys_strings(mapping: Mapping, path=None):
        if path is None:
            path = []

        non_string_keys = [x for x in mapping.keys() if not isinstance(x, str)]
        if non_string_keys:
            context_string = (
                (" in context" + ".".join(path)) if path else " in root context"
            )
            raise IOError("Non-string key(s) " + str(non_string_keys) + context_string)

        for val in mapping.values():
            if isinstance(val, Mapping):
                YAMLParametersLoader._check_all_keys_strings(val)

    _INTERPOLATION_REGEX = re.compile(r"%([\w\.]+)%")

    # noinspection PyProtectedMember
    @staticmethod
    def _interpolate(to_interpolate: Parameters, context: Parameters) -> Parameters:
        """Perform interpolation within arbitrarily nested Parameters, looking up values
        in a context if necessary.

        Limitations:
        - There cannot be interpolation in keys.
        """
        # In order to perform arbitrary interpolation, we perform topological sort on a graph where
        # the nodes are parameter keys, and edges point from those keys to each interpolation group
        # in that key's value. For example, the parameter entry `foo: %bar%/projects/%meep.baz%`
        # would give the edges (foo, bar) and (foo, meep.baz).
        #
        # pylint:disable=protected-access
        nodes = list(to_interpolate._data.keys())
        edges = list()
        for key, val in to_interpolate._data.items():
            if isinstance(val, str):
                for interp_match in YAMLParametersLoader._INTERPOLATION_REGEX.findall(
                    val
                ):
                    nodes.append(interp_match)
                    edges.append((key, interp_match))
        g = Digraph(nodes=nodes, edges=edges)
        # Since each edge has been created to point from a key to a dependency, this is to make the
        # ordering start from the leaves.
        interpolation_ordering = tuple(reversed(tuple(g.topological_sort())))

        # Perform the interpolation in-place.
        interpolated_mapping = dict(to_interpolate._data)

        for param_to_interpolate in interpolation_ordering:
            # first, we need to get the *uninterpolated* parameters value
            # (i.e. with %foo%s still present)
            if param_to_interpolate in interpolated_mapping:
                param_value_prior_to_interpolation = interpolated_mapping[
                    param_to_interpolate
                ]
            else:
                # if the parameter is not present in the parameters we are interpolating directly,
                # we look it up in the context.
                try:
                    param_value_prior_to_interpolation = context._private_get(
                        param_to_interpolate
                    )
                except ParameterError:
                    raise ParameterError(
                        f"The key '{param_to_interpolate}' doesn't exist in the parameters."
                    )

            # next, we need to actually interpolate the values
            if isinstance(param_value_prior_to_interpolation, str):
                # if this is interpolated to be a Parameters, what should its namespace prefix be?
                sub_params_namespace_prefix = list(to_interpolate.namespace_prefix)
                sub_params_namespace_prefix.append(param_to_interpolate)

                param_value_after_interpolation = _recursively_replace_matches(
                    param_value_prior_to_interpolation,
                    YAMLParametersLoader._INTERPOLATION_REGEX,
                    interpolated_mapping,
                    namespace_prefix=sub_params_namespace_prefix,
                )
            elif isinstance(param_value_prior_to_interpolation, Parameters):
                param_value_after_interpolation = YAMLParametersLoader._interpolate(
                    param_value_prior_to_interpolation, context
                )
            else:
                # we don't know how to do any interpolation for non-strings
                param_value_after_interpolation = param_value_prior_to_interpolation
            interpolated_mapping[param_to_interpolate] = param_value_after_interpolation
        return Parameters.from_mapping(
            immutabledict(
                (key, interpolated_mapping[key]) for key in to_interpolate._data.keys()
            ),
            namespace_prefix=to_interpolate.namespace_prefix,
        )

    def _unify(
        self,
        old: Parameters,
        new: Parameters,
        *,
        namespace_prefix: Sequence[str] = tuple(),
    ) -> Parameters:
        # pylint:disable=protected-access
        ret = dict()
        for (key, old_val) in old._data.items():
            if key in new:
                new_val = new._data[key]
                if isinstance(old_val, Parameters) != isinstance(new_val, Parameters):
                    if namespace_prefix:
                        namespace_prefix_str = ".".join(namespace_prefix)
                        param_str = f"{namespace_prefix_str}.{key}"
                    else:
                        param_str = key

                    raise IOError(
                        f"When unifying parameters, {param_str} is a parameter on one side and a "
                        f"namespace on the other"
                    )
                elif isinstance(old_val, Parameters):
                    new_namespace_prefix = list(namespace_prefix)
                    new_namespace_prefix.append(key)
                    ret[key] = self._unify(
                        old_val, new_val, namespace_prefix=new_namespace_prefix
                    )
                else:
                    ret[key] = new_val
            else:
                ret[key] = old_val

        for (key, new_val) in new._data.items():
            if key not in old:
                ret[key] = new_val

        return Parameters.from_mapping(ret, namespace_prefix=namespace_prefix)


@attrs(frozen=True)
class YAMLParametersWriter:
    def write(self, params: Parameters, sink: Union[Path, str, CharSink]) -> None:
        # pylint:disable=protected-access

        def dictify(data):
            if isinstance(data, ImmutableDict):
                return {k: dictify(v) for (k, v) in data.items()}
            elif isinstance(data, Parameters):
                return dictify(data._data)
            else:
                return data

        if isinstance(sink, Path) or isinstance(sink, str):
            sink = CharSink.to_file(sink)
        with sink.open() as out:
            yaml.dump(
                dictify(params._data),
                out,
                # prevents leaf dictionaries from being written in the
                # human unfriendly compact style
                default_flow_style=False,
                indent=4,
                width=78,
            )


def _recursively_replace_matches(
    candidate: Any,
    pattern: Pattern[str],
    mapping: Mapping[str, Any],
    *,
    namespace_prefix: Sequence[str],
) -> Any:
    """Replace as many groups for interpolation in a string as possible.

    For a candidate that may need to be interpolated,
    - If we've reached something that isn't a string, we're done.
    - If the interpolation pattern doesn't match, we're done.
    """
    if not isinstance(candidate, str):
        return candidate
    match = pattern.search(candidate)
    if not match:
        return candidate
    variable_to_interpolate = match.group()[1:-1]
    value_to_interpolate = mapping[variable_to_interpolate]
    if isinstance(value_to_interpolate, str):
        # we call ourselves again because there may be more variables to interpolate
        return _recursively_replace_matches(
            candidate.replace(f"%{variable_to_interpolate}%", value_to_interpolate),
            pattern,
            mapping,
            namespace_prefix=namespace_prefix,
        )
    else:
        if match.span(0) == (0, len(candidate)):
            if isinstance(value_to_interpolate, Parameters):
                # These parameters probably had their own namespace prefix for where they
                # originated but they need to have the namespace prefix for
                # their interpolated location.
                return evolve(value_to_interpolate, namespace_prefix=namespace_prefix)
            else:
                return value_to_interpolate
        else:
            raise ParameterError(
                f"Can only replace an interpolation variable with a non-string "
                f"value if the variable is the entire non-interpolated "
                f"parameter value: {namespace_prefix}"
            )



from typing import Iterable, Sized, TypeVar

from vistautils.misc_utils import str_list_limited

T = TypeVar("T")


def get_only(seq: Iterable[T]) -> T:
    """
    Get the only element of a sequence or raise an `ValueError`
    """
    it = iter(seq)
    try:
        first_element = it.__next__()
        # we use the sentinel approach rather than the usual (evil) Python "attempt can catch the
        # exception" approach to avoid raising zillions of spurious exceptions on the expected
        # code path, which makes debugging a pain
        sentinel = object()
        second_element = next(it, sentinel)
        if second_element is sentinel:
            return first_element
        else:
            got_msg: str
            if isinstance(seq, Sized):
                got_msg = str_list_limited(seq, limit=10)
            else:
                got_msg = f"{first_element!r}, {second_element!r}, and possibly more."
            raise ValueError(f"Expected one item in sequence but got {got_msg}")
    except StopIteration:
        raise ValueError("Expected one item in sequence but got none")



from typing import Any


def fully_qualified_name(clazz: type) -> str:
    """
    Gets the fully-qualified name of a class.

    This is the package name, plus the module name, plus the class name (including parent
    classes for nested classes), joined by `.`s.  For builtin types or module-less types,
    the package and module portions are omitted.

    This implementation is indebted to https://stackoverflow.com/a/13653312/413345
    """
    module = clazz.__module__
    # we compared to str.__class__.__module__ so that we don't include the
    # "builtin." prefix for built-in types
    if module is None or module == str.__class__.__module__:
        return clazz.__qualname__
    return module + "." + clazz.__qualname__


def fully_qualified_name_of_type(obj: Any) -> str:
    """
    Gets the fully-qualified name of the type of this object.

    This is the package name, plus the module name, plus the class name (including parent
    classes for nested classes), joined by `.`s.  For builtin types or module-less types,
    the package and module portions are omitted.

    This implementation is indebted to https://stackoverflow.com/a/13653312/413345
    """
    return fully_qualified_name(obj.__class__)



import gzip
import io
import os
import tarfile
import types
from abc import ABCMeta, abstractmethod
from io import BytesIO
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    AnyStr,
    BinaryIO,
    Callable,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    TextIO,
    Tuple,
    Type,
    Union,
)
from zipfile import ZipFile

from attr import attrib, attrs, validators
from immutablecollections import ImmutableDict

from vistautils.misc_utils import pathify


def is_empty_directory(path: Path) -> bool:
    """
    Returns if path is a directory with no content.
    """
    sentinel = object()
    return path.is_dir() and next(path.iterdir().__iter__(), sentinel) == sentinel


class CharSource(metaclass=ABCMeta):
    """
    Something which can provide string data.

    This abstracts over whether the string data is coming from a filesystem path, a database,
    a compressed file, etc.  This should be viewed as more analogous to a `Path` object than
    to a file.

    You can get a file-like object from this by using `open` with a context manager.

    This is inspired by Guava's `CharSource`
    """

    @abstractmethod
    def open(self) -> TextIO:
        """
        Get a file-like object which reads from this CharSource.

        You should typically use this with a context manager.

        Note that if you `read` twice, the second `read` will usually (not guaranteed) return the
        same thing as the first.  If you wish to read incrementally, use `open`.
        """
        raise NotImplementedError()

    def read_all(self) -> str:
        """
        Get the entire string which can be extacted from this source.
        """
        with self.open() as file_like:
            return file_like.read()

    def readlines(self) -> List[str]:
        """
        Get the entire string which can be extracted from this source as a list of lines.

        Lines are always broken on `\n` and line endings are not removed.
        """
        with self.open() as file_like:
            return [line.rstrip("\n") for line in file_like]

    def is_empty(self) -> bool:
        """
        Gets whether this source has a non-empty string to provide.
        """
        with self.open() as inp:
            return inp.read(1) == ""

    @staticmethod
    def from_nowhere() -> "CharSource":
        """
        An empty source.
        """
        return _StringCharSource("")

    @staticmethod
    def from_string(s: str) -> "CharSource":
        """
        Get a source whose content is the given string.
        """
        return _StringCharSource(s)

    @staticmethod
    def from_byte_source(
        wrapped_source: "ByteSource", *, encoding="utf-8"
    ) -> "CharSource":
        """
        Gets a source for a ``ByteSource`` decoded according to the given encoding.

        The default encoding is UTF-8.
        """
        return _CharSourceWrappingByteSource(wrapped_source, encoding)

    @staticmethod
    def from_file(p: Union[str, Path]) -> "CharSource":
        """
        Get a source whose content is that of the given file.

        The file will be interpreted as UTF-8.
        """
        return _FileCharSource(pathify(p))

    @staticmethod
    def from_gzipped_file(p: Path, encoding: str = "utf-8") -> "CharSource":
        """
        Get a source whose content is the uncompressed content of the given file.

        Args:
            p: path to the file whose uncompressed content should be exposed
            encoding: the encoded of the uncompressed data. Defaults to UTF-8
        """
        return _GZipFileSource(p, encoding)

    @staticmethod
    def from_file_in_tgz(
        tgz_path: Path, path_within_tgz: str, encoding: str = "utf-8"
    ) -> "CharSource":
        """
        Gets a source whose content is that of the file at the given path within a .tar.gz file.

        Note that because .tar.gz files don't support random access, using this can be slow. In
        particular, avoid using this in a loop.
        """
        return _FileWithinTgzCharSource(tgz_path, path_within_tgz, encoding)

    @staticmethod
    def from_file_in_zip(
        zip_file: Union[Path, ZipFile], path_within_zip: str
    ) -> "CharSource":
        """
        Gets a source whose content is that of the file at the given path within a .zip file.

        The `zip_file` can be specified either as a ``Path`` or a ``ZipFile`` object.
        If the latter, this ``CharSource`` is only valid as long as that ``ZipFile``
        remains open.

        The file will be interpreted as UTF-8 text.
        """
        return CharSource.from_byte_source(
            ByteSource.from_file_in_zip(zip_file, path_within_zip)
        )


@attrs(slots=True, frozen=True, repr=False)
class _StringCharSource(CharSource):
    _string = attrib(validator=validators.instance_of(str))

    def open(self) -> TextIO:
        return io.StringIO(self._string)

    def is_empty(self) -> bool:
        return not self._string

    def __repr__(self) -> str:
        if len(self._string) > 100:
            s = self._string[:100] + "..."
        else:
            s = self._string
        return f"_StringCharSource({s})"


@attrs(slots=True, frozen=True)
class _FileCharSource(CharSource):
    _path = attrib(validator=validators.instance_of(Path))

    def open(self) -> TextIO:
        return open(self._path, "r", encoding="utf-8")

    def is_empty(self) -> bool:
        return os.path.getsize(self._path) == 0


@attrs(slots=True, frozen=True, auto_attribs=True)
class _CharSourceWrappingByteSource(CharSource):
    _wrapped_source: "ByteSource"
    _encoding: str

    def open(self) -> TextIO:
        return io.TextIOWrapper(self._wrapped_source.open(), encoding=self._encoding)


@attrs(slots=True, frozen=True)
class _GZipFileSource(CharSource):
    _path: Path = attrib(validator=validators.instance_of(Path))
    _encoding: str = attrib(validator=validators.instance_of(str))

    def open(self) -> TextIO:
        return gzip.open(self._path, "rt", encoding=self._encoding)  # type: ignore

    def is_empty(self) -> bool:
        with gzip.open(self._path) as inp:
            data = inp.read(1)
        return len(data) == 0


@attrs(slots=True, frozen=True)
class _FileWithinTgzCharSource(CharSource):
    _tgz_path: Path = attrib(validator=validators.instance_of(Path))
    _path_within_tgz: str = attrib(validator=validators.instance_of(str))
    _encoding: str = attrib(validator=validators.instance_of(str))

    def open(self) -> TextIO:
        tgz_file = tarfile.open(self._tgz_path, "r:gz", encoding=self._encoding)
        # extractfile here returns  binary file object. We are lazy here and load it all into
        # memory to make dealing with the encoding issues simple
        tgz_data = tgz_file.extractfile(self._path_within_tgz)
        if tgz_data is not None:
            ret = CharSource.from_string(tgz_data.read().decode(self._encoding)).open()
            # we need to fiddle with the close method on the returned TextIO so that when it is
            # closed the containing zip file is closed as well
            old_close: Callable = ret.close
        else:
            raise IOError(
                f"Could not extract path {self._path_within_tgz} from {self._tgz_path}"
            )

        def new_close(_):
            old_close()  # pylint:disable=not-callable
            tgz_file.close()

        ret.close = types.MethodType(new_close, ret)  # type: ignore
        return ret

    def is_empty(self) -> bool:
        tgz_file = tarfile.open(self._tgz_path, "r:gz", encoding=self._encoding)
        return tgz_file.getmember(self._path_within_tgz).size == 0


class CharSink(metaclass=ABCMeta):
    """
    Something which can accept string data.

    This abstracts over whether the string data is being written to a filesystem path,
    a database, a compressed file, nowhere, etc.  This should be viewed as more analogous to a
    `Path` object than to a file.

    You can get a file-like object from this by using `open` with a context manager.

    This is inspired by Guava's `CharSink`.
    """

    @abstractmethod
    def open(self) -> TextIO:
        """
        Get a file-like object which write to this sink.

        You should typically use this with a context manager
        """
        raise NotImplementedError()

    @staticmethod
    def to_nowhere() -> "CharSink":
        """
        Get a sink which ignores its input.
        """
        return _NullCharSink()

    @staticmethod
    def to_file(p: Union[Path, str]) -> "CharSink":
        """
        Get a sink which writes to the given path.

        UTF-8 encoding will be used.
        """
        return _FileCharSink(p)

    @staticmethod
    def to_string() -> "StringCharSink":
        """
        Gets a sink which writes to a string buffer.

        See 'StringCharSink' for how to retrieve what has been written.
        """
        return StringCharSink()

    def write(self, data: str) -> None:
        """
        Write the given data to the sink.

        Note that if you `write` twice, the second `write` will overwrite the first.  If you wish
        to write incrementally, use `open`.
        """
        with self.open() as out:
            out.write(data)


class ByteSource(metaclass=ABCMeta):
    """
    Something which can provide byte data.

    This abstracts over whether the byte data is coming from a filesystem path, a database,
    a compressed file, etc.  This should be viewed as more analogous to a `Path` object than
    to a file.
    """

    @abstractmethod
    def open(self) -> BinaryIO:
        """
        Get a file-like object which reads from this ByteSource.
        """
        raise NotImplementedError()

    def read(self, size=-1) -> bytes:
        """
        Read and return data from the stream.

        If size is specified, at most size bytes will be read
        """
        with self.open() as binary_like:
            return binary_like.read(size)

    @staticmethod
    def from_file(p: Path) -> "ByteSource":
        """
        Get a source whose content is that of the given file.
        """
        return _FileByteSource(p)

    @staticmethod
    def from_file_in_zip(
        zip_file: Union[Path, ZipFile], path_within_zip: str
    ) -> "ByteSource":
        """
        Gets a source whose content is that of the file at the given path within a .zip file

        The `zip_file` can be specified either as a ``Path`` or a ``ZipFile`` object.
        If the latter, this ``ByteSource`` is only valid as long as that ``ZipFile``
        remains open.
        """
        if isinstance(zip_file, ZipFile):
            return _ByteSourceFromPathInOpenZipFile(zip_file, path_within_zip)
        else:
            return _ByteSourceFromPathInZipFile(zip_file, path_within_zip)


@attrs(slots=True, frozen=True)
class _FileByteSource(ByteSource):
    _path = attrib(validator=validators.instance_of(Path))

    def open(self) -> BinaryIO:
        return open(self._path, "rb")

    def is_empty(self) -> bool:
        return os.path.getsize(self._path) == 0


@attrs(slots=True, frozen=True)
class _ByteSourceFromPathInZipFile(ByteSource):
    _zip_path = attrib(validator=validators.instance_of(Path))
    _path_within_zip = attrib(validator=validators.instance_of(str))

    def open(self) -> BytesIO:
        # pylint:disable=not-callable
        # pylint:disable=unused-argument
        zip_file = ZipFile(self._zip_path, "r")
        ret = zip_file.open(self._path_within_zip, "r")
        # we need to fiddle with the close method on the returned BytesIO so that when it is
        # closed the containing zip file is closed as well
        old_close = ret.close

        def new_close(self):
            old_close()
            zip_file.close()

        ret.close = types.MethodType(new_close, ret)  # type: ignore
        return ret  # type: ignore


@attrs(slots=True, frozen=True)
class _ByteSourceFromPathInOpenZipFile(ByteSource):
    _zip_file = attrib(validator=validators.instance_of(ZipFile))
    _path_within_zip = attrib(validator=validators.instance_of(str))

    # mypy freaks out with this open function
    def open(self) -> BytesIO:  # type: ignore
        return self._zip_file.open(self._path_within_zip, "r")  # type: ignore


class ByteSink(metaclass=ABCMeta):
    """
    Something which can accept binary data.

    This abstracts over whether the binary data is being written to a filesystem path,
    a database, a compressed file, nowhere, etc.  This should be viewed as more analogous to a
    `Path` object than to a file.

    You can get a file-like object from this by using `open` with a context manager.

    This is inspired by Guava's `ByteSink`.
    """

    @abstractmethod
    def open(self) -> BinaryIO:
        raise NotImplementedError()

    @staticmethod
    def file_in_zip(zip_file: Path, filename_in_zip: str) -> "ByteSink":
        """
        Get a sink which writes to the given path in a zip file.
        """
        return _FileInZipByteSink(zip_file, filename_in_zip)

    @staticmethod
    def to_buffer() -> "BufferByteSink":
        """
        Get a sink which writes to a buffer in memory.

        The last bytes written can be read using methods on `BufferByteSink`
        """
        return BufferByteSink()

    @staticmethod
    def to_file(path: Union[str, Path]) -> "ByteSink":
        """
        Get a sink which writes to the given file.
        """
        return _FileByteSink(path)

    def write(self, data: bytes) -> None:
        """
        Write the given data to the sink.

        Note that if you `write` twice, the second `write` will overwrite the first.
        If you wish to write incrementally, use `open`.
        """
        with self.open() as out:
            out.write(data)


@attrs(slots=True, frozen=True)
class _NullCharSink(CharSink):
    """
    A `CharSink` which throws away anything sent to it.
    """

    def open(self) -> TextIO:
        return _NullCharSink.NullFileLike()

    class NullFileLike(TextIO):
        def __enter__(self) -> TextIO:
            return self

        def name(self):
            return "NullCharSink"

        def mode(self):
            return "w"

        def closed(self):
            return False

        def buffer(self) -> BinaryIO:
            raise NotImplementedError(
                "This isn't supposed to be part of the TextIO API"
                " but the type-checker requires it"
            )

        def encoding(self) -> str:
            return "utf-8"

        def errors(self) -> Optional[str]:
            return None

        def line_buffering(self) -> bool:
            return False

        def newlines(self) -> Any:
            return "\n"

        def close(self) -> None:
            pass

        def fileno(self) -> int:
            pass

        def flush(self) -> None:
            pass

        def isatty(self) -> bool:
            return False

        def read(self, n: int = 0) -> AnyStr:
            raise NotImplementedError("Null char sink is write-only")

        def readable(self) -> bool:
            raise NotImplementedError("Null char sink is write-only")

        def readline(self, limit: int = 0) -> AnyStr:
            raise NotImplementedError("Null char sink is write-only")

        def readlines(self, hint: int = 0) -> List[AnyStr]:
            raise NotImplementedError("Null char sink is write-only")

        def seek(self, offset: int, whence: int = 0) -> int:
            raise NotImplementedError("Null char sink is write-only")

        def seekable(self) -> bool:
            return False

        def tell(self) -> int:
            pass

        def truncate(self, size: Optional[int] = 0) -> int:
            raise NotImplementedError("Cannot truncate null char sink")

        def writable(self) -> bool:
            return True

        def write(self, s: AnyStr) -> int:
            pass

        def writelines(self, lines: Iterable[AnyStr]) -> None:
            pass

        def __next__(self) -> AnyStr:
            raise NotImplementedError("Null char sink is write-only")

        def __iter__(self) -> Iterator[AnyStr]:
            raise NotImplementedError("Null char sink is write-only")

        def __exit__(
            self,
            t: Optional[Type[BaseException]],
            value: Optional[BaseException],
            traceback: Optional[TracebackType],
        ) -> bool:
            pass


class StringCharSink(CharSink):
    """
    A sink which writes to a string buffer.

    The last string written can be recovered from the 'last_string_written' field.
    """

    def __init__(self):
        self.last_string_written = None

    def open(self) -> TextIO:
        outer_self = self

        class StringFileLike(io.StringIO):
            def __exit__(self, exc_type, exc_val, exc_tb):
                outer_self.last_string_written = self.getvalue()
                super().__exit__(exc_type, exc_val, exc_tb)

        return StringFileLike()


@attrs(slots=True, frozen=True)
class _FileCharSink(CharSink):
    _path: Union[Path, str] = attrib(validator=validators.instance_of(tuple((Path, str))))

    def open(self) -> TextIO:
        return open(self._path, "w")


@attrs(slots=True, frozen=True)
class _FileByteSink(ByteSink):
    _path: Union[Path, str] = attrib(validator=validators.instance_of(tuple((Path, str))))

    def open(self) -> BinaryIO:
        return open(self._path, "wb")


@attrs(slots=True, frozen=True)
class _FileInZipByteSink(ByteSink):
    _zip_path = attrib(validator=validators.instance_of(Path))
    _path_within_zip = attrib(validator=validators.instance_of(str))

    def open(self) -> BytesIO:
        # pylint:disable=not-callable
        # pylint:disable=unused-argument
        zip_file = ZipFile(self._zip_path, "a")
        ret = zip_file.open(self._path_within_zip, "w")
        # we need to fiddle with the close method on the returned BytesIO so that when it is
        # closed the containing zip file is closed as well
        old_close = ret.close

        def new_close(self):
            old_close()
            zip_file.close()

        ret.close = types.MethodType(new_close, ret)  # type: ignore
        return ret  # type: ignore


class BufferByteSink(ByteSink):
    """
    A sink which writes to a byte buffer.

    The last byte string written can be recovered from the 'last_bytes_written' field.
    """

    def __init__(self):
        self.last_bytes_written: bytes = None

    def open(self) -> BytesIO:
        outer_self = self

        class BytesFileLike(io.BytesIO):
            def __exit__(self, exc_type, exc_val, exc_tb):
                outer_self.last_bytes_written = self.getvalue()
                super().__exit__(exc_type, exc_val, exc_tb)

        return BytesFileLike()


def write_doc_id_to_file_map(
    doc_id_to_file_map: Mapping[str, Path], sink: CharSink
) -> None:
    """
    Writes a tab-separated docID-to-file-map to the specified sink.
    """
    with sink.open() as out:
        for doc_id in sorted(doc_id_to_file_map.keys()):
            out.write(
                "{!s}\t{!s}\n".format(doc_id, doc_id_to_file_map[doc_id].absolute())
            )


def read_doc_id_to_file_map(source: CharSource) -> Mapping[str, Path]:
    """
    Read a tab-separate docID-to-file map from the specified source.
    """
    items: List[Tuple[str, Path]] = []
    with source.open() as inp:
        for (line_num, line) in enumerate(inp):
            if line:
                parts = line.split("\t")
                if len(parts) == 2:
                    items.append((parts[0].strip(), Path(parts[1].strip())))
                else:
                    raise IOError(
                        "Bad docID to file map line {!s}: {!s}".format(line_num, line)
                    )
    return ImmutableDict.of(items)



import warnings
from functools import partial
from typing import Any, Callable, Sized, Tuple, Type, Union

import immutablecollections
from attr import Factory, attrib, validators

import vistautils.preconditions


def attrib_instance_of(type_: Union[Type, Tuple[Type, ...]], *args, **kwargs):
    warnings.warn(
        "Deprecated, use attrib(validator=instance_of(type))", DeprecationWarning
    )
    # Mypy does not understand these arguments
    return attrib(  # type: ignore
        validator=validators.instance_of(type_), *args, **kwargs
    )


# TODO cannot currently be used with additional validators:
# https://github.com/isi-nlp/isi-flexnlp/issues/188
def attrib_opt_instance_of(
    type_: Union[Type, Tuple[Type, ...]], *args, default=None, **kwargs
):
    warnings.warn(
        "Deprecated, use attrib(default=<default>, validator=optional(instance_of(<type>)))",
        DeprecationWarning,
    )
    # Mypy does not understand these arguments
    return attrib(  # type: ignore
        validator=opt_instance_of(type_), default=default, *args, **kwargs
    )


def attrib_factory(factory: Callable, *args, **kwargs):
    warnings.warn("Deprecated, use attrib(factory=<factory>)", DeprecationWarning)
    # Mypy does not understand these arguments
    return attrib(default=Factory(factory), *args, **kwargs)  # type: ignore


def attrib_immutable(
    type_: Type[immutablecollections.ImmutableCollection], *args, **kwargs
):
    warnings.warn(
        "Deprecated, use attrib(converter=<collection factory>)", DeprecationWarning
    )
    _check_immutable_collection(type_)
    # Mypy does not understand these arguments
    return attrib(converter=type_.of, *args, **kwargs)  # type: ignore


def attrib_private_immutable_builder(
    type_: Type[immutablecollections.ImmutableCollection], *args, **kwargs
):
    """
    Create an immutable collection builder private attribute.

    This is called "private" because it will not appear as a constructor argument.
    """
    warnings.warn(
        "Deprecated with no direct replacement as the builders are also deprecated. "
        "Instead, create an empty private list using attrib(factory=list, init=False) "
        "and append items to be passed to the immutablecollections factory function.",
        DeprecationWarning,
    )
    _check_immutable_collection(type_)
    # Mypy does not understand these arguments
    return attrib(  # type: ignore
        default=Factory(type_.builder), init=False, *args, **kwargs
    )


# TODO: The use of Type[ImmutableCollection] causes Mypy warnings
# Perhaps the solution is to make ImmutableCollection a Protocol?
def attrib_opt_immutable(
    type_: Type[immutablecollections.ImmutableCollection], *args, **kwargs
):
    """Return a attrib with a converter for optional collections.

    The returned attrib will create an empty collection of the
    specified type if either the attribute is not specified or it is
    specified with the value of None. Handling None allows for easier
    handling of arguments with a default absent value.
    """
    warnings.warn(
        "Deprecated, use attrib(converter=<collection factory>, default=<collection factory>)",
        DeprecationWarning,
    )
    _check_immutable_collection(type_)
    # Mypy does not understand these arguments
    return attrib(  # type: ignore
        converter=partial(_empty_immutable_if_none, type_=type_),
        default=type_.empty(),
        *args,
        **kwargs
    )


def opt_instance_of(type_: Union[Type, Tuple[Type, ...]]) -> Callable:
    warnings.warn("Deprecated, use optional(instance_of(<type>))", DeprecationWarning)
    # Mypy does not understand these arguments
    return validators.instance_of((type_, type(None)))  # type: ignore


def _check_immutable_collection(type_):
    vistautils.preconditions.check_arg(
        issubclass(type_, immutablecollections.ImmutableCollection),
        "Type {} is not an immutable collection".format(type_),
    )


def _empty_immutable_if_none(
    val: Any, type_: Type[immutablecollections.ImmutableCollection]
) -> immutablecollections.ImmutableCollection:
    if val is None:
        return type_.empty()
    else:
        return type_.of(val)


# Unused arguments are to match the attrs validator signature
# noinspection PyUnusedLocal
def non_empty(_self: Any, _attr: Any, val: Sized) -> bool:
    return len(val) > 0



# suppress lots of spurious pylint errors due to @overloads
# pylint:disable=unused-argument
# pylint:disable=function-redefined
import itertools
from collections import deque
from itertools import islice
from typing import Generic, Iterable, Iterator, Optional, Tuple, TypeVar, Union, overload

from attr import attrs

from vistautils.preconditions import check_arg

_T = TypeVar("_T")


# drop will return an Iterable if it receives an Iterable and an Iterator if it receives an Iterator
# see PEP 484: https://www.python.org/dev/peps/pep-0484/#function-method-overloading


@overload
def drop(it: Iterator[_T], num_to_skip: int) -> Iterator[_T]:
    raise NotImplementedError(
        "This should be impossible to call; this function definition is for the type checker only"
    )


@overload  # noqa: F811
def drop(it: Iterable[_T], num_to_skip: int) -> Iterable[_T]:
    raise NotImplementedError(
        "This should be impossible to call; this function definition is for the type checker only"
    )


def drop(it, num_to_skip: int):  # noqa: F811
    """
    Skip `num_to_skip` elements of an ``Iterable`` or ``Iterator``.

    If `it` is an ``Iterator``, makes an ``Iterator`` which returns elements of the original
    iterator starting from the `num_to_skip+1`th element.

    If `it` is an ``Iterable``, makes an ``Iterable`` which returns elements of the original
    iterable starting from the `num_to_skip+1`th element.

    `num_to_skip` must be non-negative or a `ValueError` will be raised.
    """
    check_arg(
        num_to_skip >= 0,
        "Number of items to skip must be positive but got %s",
        (num_to_skip,),
    )
    if hasattr(it, "__next__"):
        return itertools.islice(it, num_to_skip, None)
    else:
        return _DropIterable(it, num_to_skip)


# used by `only`
_SENTINEL = object()


def only(it: Union[Iterator[_T], Iterable[_T]]) -> _T:
    """
    Get the only element in `iterable` or throw an exception.

    If `iterable` has exactly one element, return that element.
    If it has zero or more than one, a ``LookupError`` will be raised.
    """
    if hasattr(it, "__next__"):
        # noinspection PyTypeHints
        iterator: Iterator[_T] = it  # type: ignore
    else:
        iterator = iter(it)

    try:
        ret = next(iterator)
    except StopIteration:
        raise ValueError("Expected only a single element in an iterable, but got none")

    second_element = next(iterator, _SENTINEL)
    if second_element != _SENTINEL:
        raise ValueError("Expected only a single element in iterable, but got at least 2")
    return ret


@overload
def windowed(
    it: Iterator[_T], window_size: int, *, partial_windows: bool = False
) -> Iterator[Tuple[_T, ...]]:
    raise NotImplementedError(
        "This should be impossible to call; this function definition is for the type checker only"
    )


@overload  # noqa: F811
def windowed(
    it: Iterable[_T], window_size: int, *, partial_windows: bool = False
) -> Iterable[Tuple[_T, ...]]:
    raise NotImplementedError(
        "This should be impossible to call; this function definition is for the type checker only"
    )


def windowed(it, window_size: int, *, partial_windows: bool = False):  # noqa: F811
    check_arg(window_size >= 1)

    if not hasattr(it, "__next__"):
        return _WindowedIterable(
            wrapped_iterable=it, window_size=window_size, partial_windows=partial_windows
        )

    # we know at this point that it is an Iterable, but mypy might not, hence the ignores below
    if partial_windows:
        return _possibly_incomplete_windows(it, window_size)
    else:
        return _complete_windows(it, window_size)


# implementation helpers for drop()


@attrs(auto_attribs=True)
class _DropIterable(Generic[_T], Iterable[_T]):
    # we dispense with type and sign checks here since they already happened in drop
    _wrapped_iterable: Iterable[_T]
    _num_to_drop: int

    def __iter__(self) -> Iterator[_T]:
        return drop(iter(self._wrapped_iterable), self._num_to_drop)


# implementation helpers for windowed()


@attrs(auto_attribs=True)
class _WindowedIterable(Generic[_T], Iterable[Tuple[_T, ...]]):
    # we dispense with checks on these fields because `windowed` already handles it
    _wrapped_iterable: Iterable[_T]
    _window_size: int
    _partial_windows: bool

    def __iter__(self) -> Iterator[Tuple[_T, ...]]:
        return windowed(
            iter(self._wrapped_iterable),
            self._window_size,
            partial_windows=self._partial_windows,
        )


def _complete_windows(it: Iterator[_T], window_size: int) -> Iterator[Tuple[_T, ...]]:
    """
    Complete sliding windows of the given size over an iterable.

    For a sequence ``(x_0, x_1, ...., x_n)`` and window size ``w``, this will return
    the sequence ``(x_0, ..., x_{0+w}), (x_1, ..., x_{1+w}), ...``. This will *not* return
    incomplete windows at the end of an iterable (that is, if ``x_{i+w}`` would be outside
    the input sequence).

    Implementation borrowed from Jerry Kindall and ShadowRange(?) at
    https://stackoverflow.com/a/40937024/413345
    """
    win = deque(islice(it, window_size), window_size)
    if len(win) < window_size:
        return
    # cache method access for slight speed boost
    append = win.append
    yield tuple(win)
    for e in it:
        append(e)
        yield tuple(win)


def _possibly_incomplete_windows(
    it: Iterator[_T], window_size: int
) -> Iterator[Tuple[_T, ...]]:
    """
    All sliding windows of the given size over an iterable.

    For a sequence ``(x_0, x_1, ...., x_n)`` and window size ``w``, this will return
    the sequence ``(x_0, ..., x_{0+w}), (x_1, ..., x_{1+w}), ...``. This *will* return
    incomplete windows at the end of an iterable (that is, if ``x_{i+w}`` would be outside
    the input sequence); the positions in a window beyond an iterable will be filled with
    ``fill_value``.

    Adapted from complete windowing implementation borrowed from Jerry Kindall and ShadowRange(?) at
    https://stackoverflow.com/a/40937024/413345
    """
    win = deque(islice(it, window_size), window_size)
    if not win:
        return
    # cache method access for slight speed boost
    append = win.append
    yield tuple(win)
    for e in it:
        append(e)
        yield tuple(win)
    # add incomplete windows at the end
    popleft = win.popleft
    for _ in range(window_size - 1):
        popleft()
        if win:
            yield tuple(win)
        else:
            # if the window size exceeds the sequence size, we need to stop popping early
            # or we will have a bunch of empty tuples at the end
            break


def non_none(iterable: Iterable[Optional[_T]]) -> Iterable[_T]:
    """
    Make an iterator which contains all elements from *iterable* which are not *None*.
    """
    return (x for x in iterable if x is not None)
